from Crypto.PublicKey import RSA
import os
from maica.maica_utils import get_inner_path, sync_messenger, MsgType

def get_keys_path():
    prv_path = get_inner_path('keys/prv.key')
    pub_path = get_inner_path('keys/pub.key')
    prv_path = os.path.abspath(prv_path)
    pub_path = os.path.abspath(pub_path)
    return prv_path, pub_path

def export_keys(to: str):
    prv_path, pub_path = get_keys_path()
    if not (os.path.isfile(prv_path) and os.path.isfile(pub_path)):
        sync_messenger(info=f'[maica-keys] {prv_path} and {pub_path} must exist, quitting...', type=MsgType.WARN)
        return
    try:
        with open(prv_path, "r") as prvkey_file:
            prvkey = prvkey_file.read()
        with open(pub_path, "r") as pubkey_file:
            pubkey = pubkey_file.read()
        
        sync_messenger(info="[maica-keys] Keys exist, exporting...", type=MsgType.DEBUG)
        
        prv_new = os.path.abspath(os.path.join(to, 'prv.key'))
        pub_new = os.path.abspath(os.path.join(to, 'pub.key'))

        yank_file(prv_new); yank_file(pub_new)

        with open(prv_new, "x") as prvkey_file:
            prvkey_file.write(prvkey)
        with open(pub_new, "x") as pubkey_file:
            pubkey_file.write(pubkey)

        sync_messenger(info=f"[maica-keys] Keys exported to {prv_new} and {pub_new}", type=MsgType.LOG)
        return

    except Exception as e:
        sync_messenger(info=f"[maica-keys] Error: {str(e)}", type=MsgType.ERROR)

def import_keys(frm: str):
    prv_new = os.path.abspath(os.path.join(frm, 'prv.key'))
    pub_new = os.path.abspath(os.path.join(frm, 'pub.key'))
    if not (os.path.isfile(prv_new) and os.path.isfile(pub_new)):
        sync_messenger(info=f'[maica-keys] {prv_new} and {pub_new} must exist, quitting...', type=MsgType.WARN)
        return
    try:
        with open(prv_new, "r") as prvkey_file:
            prvkey = prvkey_file.read()
        with open(pub_new, "r") as pubkey_file:
            pubkey = pubkey_file.read()
        
        sync_messenger(info="[maica-keys] Keys exist, importing...", type=MsgType.DEBUG)

        prv_path, pub_path = get_keys_path()

        yank_file(prv_path), yank_file(pub_path)

        with open(prv_path, "x") as prvkey_file:
            prvkey_file.write(prvkey)
        with open(pub_path, "x") as pubkey_file:
            pubkey_file.write(pubkey)

        sync_messenger(info=f"[maica-keys] Keys exported to {prv_path} and {prv_path}", type=MsgType.LOG)
        return

    except Exception as e:
        sync_messenger(info=f"[maica-keys] Error: {str(e)}", type=MsgType.ERROR)

def yank_file(path: str, prefix="[maica-keys] "):
    if os.path.isfile(path):
        yanked_path = path + ".old"
        sync_messenger(info=f"{prefix}Target file exists already, moving it to {yanked_path}...")

        # We only keep one old copy. You should have copied them away if they're indeed important
        os.replace(path, yanked_path)

def generate_rsa_keys():
    prv_path, pub_path = get_keys_path()
    try:
        with open(prv_path, "r") as prvkey_file:
            prvkey = prvkey_file.read()
        with open(pub_path, "r") as pubkey_file:
            pubkey = pubkey_file.read()
        
        sync_messenger(info="[maica-keys] Keys exist already, skipping...", type=MsgType.WARN)
        return
    
    except Exception:
        sync_messenger(info="[maica-keys] Keys not exist, creating...", type=MsgType.DEBUG)
        os.makedirs(get_inner_path('keys'))
    
        key = RSA.generate(2048)
        
        private_key = key.export_key()

        with open(prv_path, "wb") as prv_file:
            prv_file.write(private_key)
        
        public_key = key.publickey().export_key()

        with open(pub_path, "wb") as pub_file:
            pub_file.write(public_key)

        sync_messenger(info="[maica-keys] Keys generated successfully, store with care!", type=MsgType.LOG)

# 使用示例
if __name__ == "__main__":
    generate_rsa_keys()  # 使用默认路径



"""Import layer 4"""
from typing import *
from dataclasses import dataclass
from websockets import ServerConnection
from Crypto.Random import random as crandom
from maica.maica_utils import *
from .setting_utils import MaicaSettings
from .fsc_early import RealtimeSocketsContainer, TracerayId
from .connection_utils import DbPoolManager, AiConnectionManager

def create_slink(name: str, inner: str):
    """Just like a symlink."""
    def getter(self):
        return getattr(getattr(self, inner), name)
    def setter(self, v: any):
        setattr(getattr(self, inner), name, v)
    return property(getter, setter)

@dataclass
class ConnSocketsContainer():
    """Why so many connections!!!"""
    auth_pool: Optional[DbPoolManager]=None
    maica_pool: Optional[DbPoolManager]=None
    mcore_conn: Optional[AiConnectionManager]=None
    mfocus_conn: Optional[AiConnectionManager]=None
    mvista_conn: Optional[AiConnectionManager]=None
    mnerve_conn: Optional[AiConnectionManager]=None

    def spawn_sub(self, rsc=None):
        """Spawns a per-user sub instance."""
        sub_kwargs = {k: getattr(self, k).summon_sub(rsc) if getattr(self, k) else None for k in ['auth_pool', 'maica_pool', 'mcore_conn', 'mfocus_conn', 'mvista_conn', 'mnerve_conn']}
        return ConnSocketsContainer(**sub_kwargs)

@dataclass
class FullSocketsContainer():
    """For all convenience consideration."""
    rsc: Optional[RealtimeSocketsContainer]=None
    csc: Optional[ConnSocketsContainer]=None

    def __post_init__(self):
        if not self.rsc:
            self.rsc = RealtimeSocketsContainer()
        if not self.csc:
            self.csc = ConnSocketsContainer()

    rsc_proxied = ['websocket', 'traceray_id', 'maica_settings']
    csc_proxied = ['auth_pool', 'maica_pool', 'mcore_conn', 'mfocus_conn', 'mvista_conn', 'mnerve_conn']

    # websocket: ClassVar[Optional[ServerConnection]] = None
    # traceray_id: ClassVar[TracerayId] = None
    # maica_settings: ClassVar[MaicaSettings] = None
    # auth_pool: ClassVar[Optional[AiConnectionManager]] = None
    # maica_pool: ClassVar[Optional[AiConnectionManager]] = None
    # mcore_conn: ClassVar[Optional[AiConnectionManager]] = None
    # mfocus_conn: ClassVar[Optional[AiConnectionManager]] = None
    # mvista_conn: ClassVar[Optional[AiConnectionManager]] = None
    # mnerve_conn: ClassVar[Optional[AiConnectionManager]] = None

    def __getattr__(self, k):
        if k in self.rsc_proxied:
            return getattr(self.rsc, k)
        elif k in self.csc_proxied:
            return getattr(self.csc, k)
        else:
            return super().__getattr__(k)

    def __setattr__(self, k, v):
        if k in self.rsc_proxied:
            setattr(self.rsc, k, v)
        elif k in self.csc_proxied:
            setattr(self.csc, k, v)
        else:
            super().__setattr__(k, v)



import asyncio
import re
from typeguard import check_type
from typing import *
from .post_proc import post_proc
from maica.maica_utils import *

pattern_all_punc = re.compile(r'[.。!！?？；;，,—~-]+')
pattern_uncrit_punc = re.compile(r'[.。!！?？；;，,~]+')
pattern_subcrit_punc = re.compile(r'[.。!！?？；;~]+')
pattern_crit_punc = re.compile(r'[.。!！?？~]+')
pattern_excrit_punc = re.compile(r'[!！~]+')
pattern_numeric = re.compile(r'[0-9]')
pattern_content = re.compile(r'[一-龥A-Za-z]')
pattern_semileft = re.compile(r'[(（\[]')
pattern_semiright = re.compile(r'[)）\]]')

list_uncrit_punc = list('.。!！?？；;，,~')
list_subcrit_punc = list('.。!！?？；;~')
list_crit_punc = list('.。!！?？~')
list_excrit_puc = list('!！~')

class TalkSplitV2():
    """Transplanted from py2 fe. Terribly written."""
    def __init__(self, split_limit=180):
        self.reset()
        self._split_limit = split_limit

    def reset(self):
        self.sentence_present = ''

    def is_decimal(self, five_related_cells):
        """Not just decimal. Return True if dot shouldn't be considered as a period, vice versa."""
        def num_amount(text):
            i = 0
            for c in text:
                if c.isdigit():
                    i += 1
            return i

        if five_related_cells[2] == '.':
            cnts = len(pattern_content.findall(five_related_cells))
            if (
                # If numbers in both previous and after sections, suggesting it a decimal dot
                (num_amount(five_related_cells[0:2]) and num_amount(five_related_cells[3:5]))
                # If section has too little content, and dot is not a period after a huge number
                or (cnts<=1 and num_amount(five_related_cells[0:2]) != 2)
                # If dot belongs to an abbreviation
                or five_related_cells[1].isupper()
            ):
                return True
        return False
    
    @staticmethod
    def insert_string(original_str, insert_str, index):
        """Simple."""
        return original_str[:index] + insert_str + original_str[index:]

    def add_part(self, part):
        """Simple too."""
        self.sentence_present += part

    def split_present_sentence(self):
        """Main function."""
        apc=[]; upc=[]; spc=[]; cpc=[]; epc=[]; slc=[]; src=[]
        length_present = len(self.sentence_present.encode())

        if length_present <= 60:
            return None
    
        def get_pos_len(pos):
            """Get length of context before position."""
            sce = self.sentence_present[0:pos]
            return len(sce.encode())
        
        def check_sanity_pos(pos):
            """Checks brackets' consistency."""
            lc = rc = 0
            if slc:
                for match in slc:
                    if match.end() > pos:
                        break
                    else:
                        lc += 1
            if src:
                for match in src:
                    if match.end() > pos:
                        break
                    else:
                        rc += 1

            return True if lc <= rc else False
            
        def split_at_pos(pos):
            """Just split."""
            sce = self.sentence_present[0:pos]
            self.sentence_present = self.sentence_present[pos:]
            if len(sce) > 1 and not sce.isspace():
                return sce.lstrip()
            else:
                return None

        matches = pattern_all_punc.finditer(self.sentence_present)
        for match in matches:
            pos = match.end(); content = match.group()
            apc.append(match)
            if len(content) > 1 or not self.is_decimal(('   ' + self.sentence_present + ' ')[pos:pos+5]):
                if has_words_in(content, *list_uncrit_punc):
                    upc.append(match)
                    if has_words_in(content, *list_subcrit_punc):
                        spc.append(match)
                        if has_words_in(content, *list_crit_punc):
                            cpc.append(match)
                            if has_words_in(content, *list_excrit_puc):
                                epc.append(match)

        slc = list(pattern_semileft.finditer(self.sentence_present))
        src = list(pattern_semiright.finditer(self.sentence_present))

        # if length_present <= 60:
        #     return None
        if epc:
            for match in reversed(epc):
                if 30 <= get_pos_len(match.end()) <= self._split_limit and check_sanity_pos(match.end()):
                    return split_at_pos(match.end())
        # No epc or none fits
        if length_present <= self._split_limit - 35:
            return None
        if cpc:
            for match in reversed(cpc):
                if 30 <= get_pos_len(match.end()) <= self._split_limit and check_sanity_pos(match.end()):
                    return split_at_pos(match.end())
        # No cpc or still none fits
        if length_present <= self._split_limit - 25:
            return None
        if spc:
            for match in reversed(spc):
                if 20 <= get_pos_len(match.end()) <= self._split_limit and check_sanity_pos(match.end()):
                    return split_at_pos(match.end())
        # No spc or still none fits
        if length_present <= self._split_limit - 15:
            return None
        if upc:
            for match in reversed(upc):
                if 10 <= get_pos_len(match.end()) <= self._split_limit and check_sanity_pos(match.end()):
                    return split_at_pos(match.end())
        # No upc or still none fits
        if length_present <= self._split_limit - 5:
            return None
        if apc:
            for match in reversed(apc):
                if 3 <= get_pos_len(match.end()) <= self._split_limit and check_sanity_pos(match.end()):
                    return split_at_pos(match.end())
        # Falling back -- sanity given up
            if length_present <= self._split_limit:
                return None
            for match in reversed(apc):
                if 3 <= get_pos_len(match.end()) <= self._split_limit + 20:
                    return split_at_pos(match.end())
        return split_at_pos(get_pos_len(self._split_limit + 20))
    
    def announce_stop(self):
        """Exhausts remaining buffer."""
        sce_list = []
        res = True
        while res:
            res = self.split_present_sentence()
            if res:
                sce_list.append(res)
        if self.sentence_present and len(self.sentence_present) > 1 and not self.sentence_present.isspace():
            sce_list.append(self.sentence_present.lstrip())
        self.reset()
        return sce_list

    # def add_pauses(self, strin):
    #     """Add pauses to punctuations. Suppose we aren't using this in backend."""
    #     if not isinstance(strin, str):
    #         raise TypeError("Input should be a string, get {}".format(type(strin)))
        
    #     def get_pre_i_space(ele, list):
    #         pre_index = list.index(ele) - 1
    #         pre_pos = list[pre_index][3] if pre_index >= 0 else 0
    #         return ele[3] - pre_pos

    #     iupc=[]; icpc=[]; iepc=[]
        
    #     matches = pattern_all_punc.finditer(strin)
    #     lmatch = None; i = 0 
    #     # This is a iterator but we need a list
    #     matches = list(matches)
    #     for match in matches:
    #         # No pause needed for last punc
    #         i += 1
    #         if i == len(matches):
    #             break
    #         pos = match.end(); content = match.group()
    #         if not lmatch:
    #             preseq = strin[:match.start()]
    #         else:
    #             preseq = strin[lmatch.end():match.start()]
    #         prelen = len(preseq.encode('utf-8'))
    #         alllen = len(strin[:match.start()].encode('utf-8'))
    #         lmatch = match
    #         match_tuple_b = (pos, content, prelen, alllen)
    #         # print(('   ' + strin + ' ')[pos:pos+5])
    #         if len(content) > 1 or not self.is_decimal(('   ' + strin + ' ')[pos:pos+5]):
    #             if has_words_in(content, list_excrit_puc):
    #                 iepc.append(match_tuple_b)
    #             elif has_words_in(content, list_crit_punc):
    #                 icpc.append(match_tuple_b)
    #             elif has_words_in(content, list_uncrit_punc):
    #                 iupc.append(match_tuple_b)

    #     pending_insert = []

    #     for mb in iupc:
    #         if prelen > 80:
    #             pending_insert.append(('{w=0.3}', mb[0]))

    #     for mb in icpc:
    #         if len(mb[1]) > 1:
    #             # ellipsis?
    #             if get_pre_i_space(mb, icpc) > 30:
    #                 pending_insert.append(('{w=0.5}', mb[0]))
    #         elif get_pre_i_space(mb, icpc) > 45:
    #             pending_insert.append(('{w=0.3}', mb[0]))

    #     for mb in iepc:
    #         if prelen > 45:
    #             pending_insert.append(('{w=0.2}', mb[0]))

    #     for tup in pending_insert[::-1]:
    #         strin = self.insert_string(strin, *tup)
                
    #     return strin

class PPRTProcessor():
    """Post proc realtime processor."""
    @Decos.report_limit_warning
    def __init__(self, pprt: Union[dict, True]=True, target_lang: Literal['zh', 'en']='zh', mnerve_conn: Optional[AiConnectionManager]=None):
        self._pprt = {
            "yield_interval": [40, 20, 10, 5, 3, 1],
            "split_limit": 180,
            "correct_malform": True,
        }
        if isinstance(pprt, dict):
            check_type(pprt.get('yield_interval'), List[int])
            check_type(pprt.get('split_limit'), int)
            self._pprt.update(pprt)

        self._target_lang = target_lang
        self._mnerve_conn = mnerve_conn
        self._buffer = TalkSplitV2(self._pprt.get('split_limit'))

        self._add_counter = 0
        self._yield_counter = 0

    def _add_chunk(self, chunk: str):
        self._buffer.add_part(chunk)
        self._add_counter += 1

    def _try_yield(self):
        return self._buffer.split_present_sentence()
    
    async def store_and_split(self, chunk: str) -> Optional[str]:
        self._add_chunk(chunk)
        if self._add_counter >= sum(self._pprt['yield_interval'][:self._yield_counter + 1]):
            split = self._buffer.split_present_sentence()
            if split:
                if self._pprt.get('correct_malform'):
                    return await self._correct_malform(split)
                else:
                    return split
            else:
                self._yield_counter += 1
                return None
            
    async def _correct_malform(self, sentence: str) -> str:
        return await post_proc(sentence, self._target_lang, self._mnerve_conn)

    async def exaust_and_split(self) -> list[str]:
        splits = self._buffer.announce_stop()
        new_splits = []
        for split in splits:
            if self._pprt.get('correct_malform'):
                new_splits.append(await self._correct_malform(split))
            else:
                new_splits.append(split)
        return new_splits
    
if __name__ == "__main__":
    pprtp = PPRTProcessor()
    text = "[微笑]我觉得...年轻人喜欢喝奶茶是因为它好喝吧? [微笑]奶茶口感细腻, 甜度可调, 还可以加各种配料.[开心]很多人也把喝奶茶作为一种享受生活的方式. [微笑]当然, 奶茶也有它的营养价值. [微笑]比如牛奶富含钙和蛋白质, 茶叶则含有茶多酚和咖啡因. [开心]所以适量饮用奶茶对身体也是有好处的."
    for c in text:
        sce = asyncio.run(pprtp.store_and_split(c))
        if sce:
            print(sce)
    sces = asyncio.run(pprtp.exaust_and_split())
    for sce in sces:
        print(sce)


import json
import random
import traceback
import asyncio
import websockets
import colorama

from openai.types.chat import ChatCompletion, ChatCompletionMessage
from typing import *
from maica.mfocus.mfocus_sfe import SfPersistentManager
from maica.mtrigger.mtrigger_sfe import MtPersistentManager
from maica.mfocus.agent_modules import AgentTools
from maica.mfocus.react_trigger import react_detect
from maica.maica_utils import *
from maica.mtools import providers

class MFocusManager(AgentContextManager):
    def __init__(self, fsc: FullSocketsContainer, sf_inst: SfPersistentManager, mt_inst: Optional[MtPersistentManager]=None):
        super().__init__(fsc, sf_inst, mt_inst)        
        self.agent_tools = AgentTools(fsc, sf_inst)

    def _construct_tools(self):
        self.tools = []
        if self.mt_inst and not self.settings.temp.bypass_mt:
            trigger_list = self.mt_inst.get_valid_triggers()
        else:
            trigger_list = None

        self.tools =  [
            {
                "name": "time_acquire",
                "description": "调用该工具以获取当前时间. 只要对话关于: 时间, 问候, 三餐, 休息, 生活节律与建议等, 或你需要获取当前时间以帮助回答, 就使用此工具查询时间." if self.settings.basic.target_lang == 'zh' else "Call this tool to get the current time. Always use this tool if the conversation mentions: time, greeting, meals, sleep, rest, pace of life or related suggestions, or if you want to know the current time to make your answer.",
                "parameters": {
                    "type": "object",
                    "properties": {
                    },
                    "required": [
                    ],
                    "optional": [
                    ]
                }
            },
            {
                "name": "date_acquire",
                "description": "调用该工具以获取当前日期. 只要对话关于: 日期, 季节, 年份, 或你需要获取当前日期以帮助回答, 就使用此工具查询日期." if self.settings.basic.target_lang == 'zh' else "Call this tool to get the current date. Always use this tool if the conversation mentions: date, season, year, or if you want to know the current date to make your answer.",
                "parameters": {
                    "type": "object",
                    "properties": {
                    },
                    "required": [
                    ],
                    "optional": [
                    ]
                }
            },
            {
                "name": "weather_acquire",
                "description": "调用该工具以获取当前天气. 只要对话关于: 天气, 通勤, 户外活动, 或你需要获取当前天气以帮助回答, 就使用此工具查询天气." if self.settings.basic.target_lang == 'zh' else "Call this tool to get the current weather. Always use this tool if the conversation mentions: weather, commuting, outdoor activities, or if you need to know the current weather to make your answer.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "你需要查询天气的地理位置. 如果查询用户本地天气, 则留空." if self.settings.basic.target_lang == 'zh' else "The location which you need to get the weather of. Leave empty for the user's local weather.",
                            "example_value": "湖北武汉" if self.settings.basic.target_lang == 'zh' else "Los Angeles"
                        },
                    },
                    "required": [
                    ],
                    "optional": [
                        "location",
                    ]
                }
            },
            {
                "name": "event_acquire",
                "description": "调用该工具以查询当前或指定日期的节日或事件. 只要对话关于: 日期, 节日, 活动, 假期, 或你需要获取指定的事件以帮助回答, 就使用此工具查询节日或事件." if self.settings.basic.target_lang == 'zh' else "Call this tool to get the event or holiday of a given date or current date. Always use this tool if the conversation mentions: date, holiday, anniversary, activities, vacation, or if you need to know the specific event to make your answer.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "year": {
                            "type": "number",
                            "description": "需要查询日期的年份. 如果日期在今年, 则留空." if self.settings.basic.target_lang == 'zh' else "The year of the given date, leave empty for this year.",
                            "example_value": "2003"
                        },
                        "month": {
                            "type": "number",
                            "description": "需要查询日期的月份. 如果日期在本月, 则留空." if self.settings.basic.target_lang == 'zh' else "The month of the given date, leave empty for this month.",
                            "example_value": "6"
                        },
                        "day": {
                            "type": "number",
                            "description": "需要查询日期的日数. 如果日期在本日, 则留空." if self.settings.basic.target_lang == 'zh' else "The day of the given date, leave empty for the day today.",
                            "example_value": "26"
                        },
                    },
                    "required": [
                    ],
                    "optional": [
                        "year",
                        "month",
                        "day",
                    ]
                }
            },
            {
                "name": "persistent_acquire",
                "description": "调用该工具以查询你的角色(莫妮卡)或用户的记忆, 例如你或用户的喜好, 兴趣, 经历, 体验, 关系或个人信息. 只要你被问及你作为AI不具有的记忆, 经历, 个性, 喜好或其它事物, 就使用此工具查找记忆. 此外应注意, 若用户提及'天堂树林', '太空教室', '文学部', 及其它指代不明确的专有名词, 可先使用此工具." if self.settings.basic.target_lang == 'zh' else "Call this tool to get any additional information from your character (Monika)'s memory or user's memory, such as your or user's preferences, hobbies, experiences, appearence, relationship or personal information. Always use this tool if the conversation mentions memories, personality, preferences, hobbies, experience or any other thing you don't have as an AI. Also, if user mentioned 'Heaven Forest', 'Spaceroom', 'Literature club' or unclear proper nouns, use this tool first.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "需要从你或用户的记忆中搜索的信息, 以尽可能直接回答问题. 以第三人称方式提出问题, 不要提及'我'或'我们'." if self.settings.basic.target_lang == 'zh' else "The question you want to search from your memory or user's memory, to answer the question as directly as possible. Ask in third person perspective, do not use appellation 'I', 'me' or 'us'.",
                            "example_value": "莫妮卡喜欢吃什么?" if self.settings.basic.target_lang == 'zh' else "What does Monika like to eat?"
                        },
                    },
                    "required": [
                        "query",
                    ],
                    "optional": [
                    ]
                }
            },
        ]

        if providers.get_asearch():
            self.tools.append(
                {
                    "name": "search_internet",
                    "description": "调用该工具以搜索互联网. 如果你需要联网搜索信息以帮助回答, 则使用该工具搜索互联网. 如果该信息可以通过其它工具获取, 则优先使用其它工具." if self.settings.basic.target_lang == 'zh' else "Call this tool to search a question on the Internet. Use this tool if you need information from the Internet to make your answer. If another tool avaliable could provide the information, use that tool instead.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "query": {
                                "type": "string",
                                "description": "需要在搜索引擎中搜索的问题, 应当简洁明确." if self.settings.basic.target_lang == 'zh' else "The question needs to be searched on Google, which should be brief and clear.",
                                "example_value": "附近的餐馆" if self.settings.basic.target_lang == 'zh' else "Nearby restaurants"
                            },
                            "location_req": {
                                "type": "boolean",
                                "description": "仅当问题关于用户所处的地理位置时设为true." if self.settings.basic.target_lang == 'zh' else "Set to true only if the query is associated with the user's current location.",
                                "example_value": "true"
                            }
                        },
                        "required": [
                            "query",
                        ],
                        "optional": [
                            "location_req",
                        ]
                    }
                },
            )

        choice_list = []; choice_checklist = []
        if trigger_list and self.settings.extra.amt_aggressive:
            for trigger in trigger_list:
                match trigger.template:
                    case 'common_affection_template':
                        pass
                    case 'common_switch_template':
                        cst_temp_list = []
                        for i in trigger.exprop.item_list:
                            j = f'选择{i}' if self.settings.basic.target_lang == 'zh' else f'switch to {i}'
                            choice_checklist.append(j)
                            cst_temp_list.append(j)

                        if trigger.exprop.suggestion:
                            j = f"选择未列出的{trigger.exprop.item_name.zh}" if self.settings.basic.target_lang == 'zh' else f"Choose an unlisted {trigger.exprop.item_name.en}"
                            choice_checklist.append(j)
                            cst_temp_list.append(j)

                        cst_explaination = f"更换{trigger.exprop.item_name.zh}" if self.settings.basic.target_lang == 'zh' else f"Change {trigger.exprop.item_name.en}"
                        choice_list.append({cst_explaination: cst_temp_list})

                    case 'common_meter_template':
                        cmt_iname = f'调整{trigger.exprop.item_name.zh}' if self.settings.basic.target_lang == 'zh' else f'Adjust {trigger.exprop.item_name.en}'
                        choice_checklist.append(cmt_iname)
                        j = f"{cmt_iname}, 范围是{trigger.exprop.value_limits[0]}到{trigger.exprop.value_limits[1]}" if self.settings.basic.target_lang == 'zh' else f"{cmt_iname} within range {trigger.exprop.value_limits[0]} to {trigger.exprop.value_limits[1]}"
                        choice_list.append(j)

                    case _:
                        cc_iname = trigger.exprop.item_name.zh if self.settings.basic.target_lang == 'zh' else trigger.exprop.item_name.en
                        j = f"触发{cc_iname}" if self.settings.basic.target_lang == 'zh' else f"Trigger {cc_iname}"
                        choice_checklist.append(j)
                        choice_list.append(j)

        self.choice_list = choice_list
        self.choice_checklist = choice_checklist

        if choice_list:
            self.react_trigger_impl = "func" if self.mnerve_conn else "tool"
        else:
            self.react_trigger_impl = None

        if self.react_trigger_impl == 'tool':
            self.tools.append(
                {
                    "name": "react_trigger",
                    "description": "此工具能检验用户的特定要求可否完成. 如果用户对你提出以下类别请求: 切换(如换衣服, 换地点, 换场景), 调整(如调距离, 调亮度), 触发(如开灯, 开启某模式), 则调用该工具." if self.settings.basic.target_lang == 'zh' else "This tool can verify if user's specific request can be done. If user is requesting you to: switching(like changing your clothes, changing location, changing scene), adjusting(like adjusting distance, adjusting brightness), triggering(like turning on light, turning on some mode), use this tool.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "prediction": {
                                "type": "string",
                                "description": f"你将收到一系列可用选项. 若某选项能够满足用户的请求, 则输出该选项, 否则输出false. 以下是可用选项:\n{json.dumps(choice_list)}" if self.settings.basic.target_lang == 'zh' else f"You'll be offered a list of avaliable choices. return the corresponding choice if you think it matches the query, or return false if none matches. Avaliable choices:\n{json.dumps(choice_list)}",
                                "example_value": random.choice(choice_checklist)
                            }
                        },
                        "required": [
                            "prediction",
                        ],
                        "optional": [
                        ]
                    }
                },
            )

        if self.settings.temp.mv_imgs:
            self.tools.append(
                {
                    "name": "vista_acquire",
                    "description": "用户已上传了一到数张图片. 此工具能从用户上传的图片中获取信息, 并以文字形式返回. 如果用户要求你使用视觉或查看图片等, 则调用该工具." if self.settings.basic.target_lang == 'zh' else "User has uploaded one or several images. This tool can extract information from images user uploaded, and return in plain text. If user is requesting you to use your vision, read pictures or do related things, use this tool.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "query": {
                                "type": "string",
                                "description": "需要在图片中提取的信息, 应当简洁明确. 留空表示需要概括图片内容." if self.settings.basic.target_lang == 'zh' else "The question you want to search from the image, which should be brief and clear. Leave empty to get a brief introduction of its content.",
                                "example_value": "图中人物的衣服颜色" if self.settings.basic.target_lang == 'zh' else "Color of man on the picture's clothes"
                            }
                        },
                        "required": [
                        ],
                        "optional": [
                            "query",
                        ]
                    }
                },
            )

        if self.settings.extra.mf_aggressive:
            self.tools.append(
                {
                    "name": "conclude_information",
                    "description": "此工具用于总结你的输出. 只要你已获取了所有必要信息且准备好作答, 就在作答前调用该工具." if self.settings.basic.target_lang == 'zh' else "This tool can conclude your outputs. Always use this tool if you have acquired all necessary informations before making final answer.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "conclusion": {
                                "type": "string",
                                "description": "总结你获取的信息和作出的推理, 并整理成一个简洁的句子." if self.settings.basic.target_lang == 'zh' else "Conclude all information you have acquired and reasonings you have made into a concise sentence.",
                                "example_value": "现在是上午九点, 因此适合吃早餐, 且天气凉爽, 因此适合户外活动."  if self.settings.basic.target_lang == 'zh' else "It's 9:00 in the morning, suitable for breakfast. The weather is cool, good for exercising."
                            }
                        },
                        "required": [
                            "conclusion"
                        ],
                        "optional": [
                        ]
                    }
                },
            )
        else:
            self.tools.append(
                {
                    "name": "agent_finished",
                    "description": f"若你已调用了所有其它必要的工具, 或不需要调用任何其它工具, 则在作出最终回答之前调用此工具, 以表示调用完成." if self.settings.basic.target_lang == 'zh' else f"Call this tool after you've finished calling every other necessary tool, or if you don't need any other tool. Call this tool before making final answer, so we know toolcalling has finished.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                        },
                        "required": [
                        ],
                        "optional": [
                        ]
                    }
                },
            )

        self.tools = alt_tools(self.tools)

    async def _construct_query(self, user_input=None, tool_input=None, tool_id=None):
        await super()._construct_query(user_input, tool_input, tool_id, 'pre')
    
    async def agenting(self, query) -> str:
        try:

            # Just all the tools
            instructed_answer = {
                "time_acquire": '',
                "date_acquire": '',
                "weather_acquire": '',
                "event_acquire": '',
                "persistent_acquire": '',
                "search_internet": '',
                "react_trigger": '',
                "vista_acquire": '',
            }
            conclusion_answer = None

            def _instructed_add(tool_real_name: str, humane: Union[str, list], edit=True):
                nonlocal instructed_answer

                # We merge lists and overwrite strings by default
                if not instructed_answer.get(tool_real_name):
                    instructed_answer[tool_real_name] = humane
                elif edit:
                    if isinstance(instructed_answer[tool_real_name], list):
                        if isinstance(humane, list):
                            instructed_answer[tool_real_name].extend(humane)
                        else:
                            # Likely less valuable so we discard
                            pass
                    else:
                        # Overwrite it anyway
                        instructed_answer[tool_real_name] = humane
                # The logic is not that considerate, but trust me it's enough

            def _react_to_instruction(resp):
                nonlocal self

                if not has_valid_content(resp):
                    inst = '[player]的请求当前无法被满足. 请表示你做不到, 并建议[player]自行解决或寻找其它方法.' if self.settings.basic.target_lang == 'zh' else '[player]\'s current request cannot be satisfied. please indicate that you can\'t do it, and suggest [player] doing it themselves or find another way.'
                else:
                    if resp in self.choice_checklist:
                        inst = f'[player]的请求是你所了解的, 且会被系统完成, 请作出关于<{resp}>的正面答复.' if self.settings.basic.target_lang == 'zh' else f'[player]\'s request is understood and will be done by system, please make positive answer about <{resp}>.'
                    else:
                        inst = '[player]的请求是你所了解的, 且会被系统完成, 请作出正面答复.' if self.settings.basic.target_lang == 'zh' else '[player]\'s request is understood and will be done by system, please make positive answer.'

                return inst

            # First thing first we prepare the first query
            if self.settings.temp.mv_imgs:
                image_word = " [图片]" if self.settings.basic.target_lang == 'zh' else " [Image]"
                query += image_word
            self._construct_tools()
            await self._construct_query(user_input=query)

            if self.react_trigger_impl == 'func':
                react_task = asyncio.create_task(react_detect(query, self.fsc, self.choice_list))
            else: react_task = None

            cycle = 0; ending = False
            while cycle <= 7 and not ending:

                # Sanity check
                cycle += 1

                resp_content, resp_reasoning, resp_tools = await self._send_query(thinking=True)
                resp_content, resp_reasoning = proceed_common_text(resp_content), proceed_common_text(resp_reasoning)

                await messenger(self.websocket, 'maica_mfocus_toolchain', f'\nMFocus toolchain {cycle} round responded, response is:\nR: {resp_reasoning}\nA: {resp_content}\nAnalyzing response...', code='200')
                tool_seq = 0
                if resp_tools:
                    for resp_tool in resp_tools:

                        # Tool parallel support
                        tool_seq += 1
                        tool_id, tool_type, tool_func_name, tool_func_args = resp_tool.id, resp_tool.type, resp_tool.function.name, resp_tool.function.arguments
                        await messenger(self.websocket, 'maica_mfocus_parallel_tool', f'\nCalling parallel tool {tool_seq}/{len(resp_tools)}:\n{resp_tool}\nGathering information...', '200', type=MsgType.INFO, color=colorama.Fore.BLUE)

                        machine = humane = None
                        args = []

                        kwargs = proceed_common_text(tool_func_args, is_json=True)
                        if not kwargs:
                            kwargs = {}

                        if tool_func_name == "search_internet":
                            kwargs['original_query'] = query
                        elif tool_func_name == "vista_acquire":
                            kwargs['img_list'] = self.settings.temp.mv_imgs

                        function_route = getattr(self.agent_tools, tool_func_name, None)
                        if function_route:
                            machine, humane = await function_route(*args, **kwargs)
                        else:
                            match tool_func_name:
                                case 'react_trigger':
                                    humane = _react_to_instruction(kwargs.get('prediction'))
                                    machine = '已收到你的判断, 请继续调用其它工具或正常结束作答.' if self.settings.basic.target_lang == 'zh' else 'Your judgement recieved, please continue using other tools or end as normal.'
                                case 'conclude_information':
                                    conclusion_answer = kwargs.get('conclusion')
                                    await messenger(self.websocket, 'maica_mfocus_conclusion', f'\nMFocus conclusion recieved:\n{conclusion_answer}\nEnding toolchain...', '200', type=MsgType.INFO, color=colorama.Fore.LIGHTBLUE_EX)
                                    ending = True
                                    break
                                case 'agent_finished':
                                    await messenger(self.websocket, 'maica_mfocus_empty', f'MFocus null recieved, ending toolchain...', '204', type=MsgType.INFO, color=colorama.Fore.LIGHTBLUE_EX)
                                    ending = True
                                    break
                                case _:
                                    # This tool call is unrecognizable
                                    raise MaicaInputError('Unrecognizable toolcall recieved', '405')
                                
                        if not has_valid_content(machine):
                            await messenger(self.websocket, 'maica_mfocus_parallel_empty', f'Answer to parallel tool {tool_seq}/{len(resp_tools)} is empty', '200', type=MsgType.WARN)
                            machine = '未获得有效信息' if self.settings.basic.target_lang == 'zh' else 'No useful information found'

                        elif has_valid_content(humane):
                            await messenger(self.websocket, 'maica_mfocus_parallel_result', f'Answer to parallel tool {tool_seq}/{len(resp_tools)} is "{ellipsis_str(humane, 50)}"', '200', type=MsgType.INFO)
                            _instructed_add(tool_func_name, humane)

                        await self._construct_query(tool_input=machine, tool_id=tool_id)

                else:
                    await messenger(self.websocket, 'maica_mfocus_absent', f'No tool called, ending toolchain...', '204', type=MsgType.INFO, color=colorama.Fore.LIGHTBLUE_EX)
                    ending = True

                if self.settings.extra.pre_astp and not ending:
                    await messenger(self.websocket, 'maica_mfocus_astp', f'MFocus interrupted by pre_astp, ending toolchain...', '200', type=MsgType.INFO, color=colorama.Fore.LIGHTBLUE_EX)
                    ending = True

                await messenger(self.websocket, 'maica_mfocus_round_finish', f'MFocus toolchain {cycle} round finished, ending is {str(ending)}', '200', type=MsgType.INFO, color=colorama.Fore.BLUE)
                    
            # Now we're out of the loop

            if react_task:
                await react_task
                _instructed_add('react_trigger', _react_to_instruction(react_task.result()))

            if self.settings.extra.mf_aggressive:
                
                # So we use last response instead if no conclusion offered
                if not conclusion_answer:
                    conclusion_answer = proceed_common_text(resp_content)

                # If there is information and answer
                if cycle >= 2 and conclusion_answer:
                    await messenger(self.websocket, 'maica_mfocus_using_conclusion', 'MFocus got conclusion and used', '200')
                    return conclusion_answer
                
                await messenger(self.websocket, 'maica_mfocus_no_conclusion', 'MFocus got no conclusion, falling back to instruction', '404', traceray_id=self.traceray_id)
                
            # Then if mfa not enabled or ignored
            if self.settings.extra.tnd_aggressive >= 1:
                # Add time and events
                if not instructed_answer.get('time_acquire'):
                    _instructed_add('time_acquire', (await self.agent_tools.time_acquire())[1], False)
                if not instructed_answer.get('event_acquire'):
                    _instructed_add('event_acquire', (await self.agent_tools.event_acquire())[1], False)
            if self.settings.extra.tnd_aggressive >= 2 or self.settings.temp.ic_prep:
                # Add date and weather
                if not instructed_answer.get('date_acquire'):
                    _instructed_add('date_acquire', (await self.agent_tools.date_acquire())[1], False)
                if not instructed_answer.get('weather_acquire'):
                    _instructed_add('weather_acquire', (await self.agent_tools.weather_acquire())[1], False)

            instructed_answer_str = ''
            if instructed_answer:
                for k, v in instructed_answer.items():
                    # v can be only list or str
                    try:
                        if k in ["persistent_acquire", "search_internet"]:
                            if v:
                                v = '"' + v + '"' if isinstance(v, str) else '"' + ', '.join(v) + '"'
                            else:
                                v = ''
                        else:
                            if v:
                                v = '[' + v + ']' if isinstance(v, str) else str(v)
                            else:
                                v = ''
                        instructed_answer_str += v
                    except Exception as e:
                        await messenger(self.websocket, 'maica_mfocus_corruption', 'MFocus got corrupted information, skipping to continue', '304', traceray_id=self.traceray_id, type=MsgType.WARN)
            else:
                # If there's really no instruction
                return None

            return instructed_answer_str
        
        except CommonMaicaException as ce:
            raise ce
        
        except websockets.WebSocketException as we:
            raise MaicaConnectionWarning(str(we), '408') from we

        except Exception as e:
            raise CommonMaicaError('Uncaught MFocus exception happened', '500', 'maica_mfocus_critical') from e



import time
import asyncio
from typing import *

from maica.maica_utils import *

class BufferDict(dict):
    """Specifically used for stream buffering."""
    def add_id(self, id: int):
        """Also overwrites."""
        self[id] = StreamBuffer()

    def del_id(self, id: int):
        self.pop(id, None)

class StreamBuffer:
    """A queue-like buffer."""
    
    def __init__(self):
        self.timestamp = time.time()
        self._queue = asyncio.Queue()
        self._exhausted = asyncio.Event()
    
    async def aexhaust(self):
        self._exhausted.set()

    async def aappend(self, item):
        await self._queue.put(item)
    
    def sappend(self, item):
        try:
            self._queue.put_nowait(item)
        except asyncio.QueueFull:
            pass
    
    def __aiter__(self):
        return self
    
    async def __anext__(self):
        if not len(self) and self._exhausted.is_set():
            raise StopAsyncIteration
        
        tasks = []
        tasks.append(asyncio.create_task(self._queue.get()))
        if not self._exhausted.is_set():
            tasks.append(asyncio.create_task(self._exhausted.wait()))

        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        for future in pending:
            future.cancel()
            await future
        for future in done:
            item = future.result()
            if item == True:
                raise StopAsyncIteration
            elif item:
                return item
    
    def __iter__(self):
        while not self._queue.empty():
            try:
                yield self._queue.get_nowait()
            except asyncio.QueueEmpty:
                break
    
    def __len__(self):
        return self._queue.qsize()
    
buffer_dict: BufferDict[StreamBuffer] = BufferDict()


import re
import json
import asyncio
import traceback
import functools

from maica.maica_utils import *
from . import providers

async def internet_search(fsc: FullSocketsContainer, query, original_query):
    target_lang = fsc.maica_settings.basic.target_lang

    @Decos.conn_retryer_factory()
    async def _search(fake_self, query, target_lang):
        results_sync = await (providers.get_asearch())(query, target_lang)
        assert len(results_sync), 'Search result is empty'
        return results_sync

    results_sync = await _search(DummyClass(name="serp"), query, target_lang)

    results_full = []
    results_short = []
    results_humane = ''
    rank = 0
    for item in results_sync:
        rank += 1
        title, text = clean_text(item['title']), clean_text(item['text'])
        results_full.append({'rank': rank, 'title': title, 'text': text})
        if rank <= 5:
            results_short.append({'rank': rank, 'title': title, 'text': text})
        if rank <= 3:
            humane_text = ReUtils.re_sub_serp_datetime.sub('', text, 1)
            results_humane += f'信息{rank}\n标题:{title}\n内容:{humane_text}\n'

    await messenger(info=f'MFocus got {rank} information lines from search engine', type=MsgType.INFO)

    results_full_str = str(results_full).strip('[').strip(']')
    results_short_str = str(results_short).strip('[').strip(']')
    if not fsc.maica_settings.extra.esc_aggressive:
        return results_short_str, results_humane

    system_init = """你是一个人工智能助手, 你接下来会收到一个问题和一些来自互联网的信息.
以单行不换行的自然语言的形式, 简洁地整理提供相关的信息, 长度不要超过一个自然句. 如果你最终认为提供的信息与问题相关性不足, 返回false.
Begin!""" if target_lang == 'zh' else """You are a helpful assistant, now you will recieve a question and some information from the Internet.
Conclude related information briefly in a single line of natural language, and do not exceed the length of a natural sentence. If you think the provided information is not related enough to the question, return false.
Begin!"""

    messages = [{'role': 'system', 'content': system_init}]
    messages.append({'role': 'user', 'content': f'question: {query}; information: {results_full_str}'})
    # messages = apply_postfix(messages, thinking=False)
    completion_args = {
        "messages": messages,
    }

    conn = fsc.mnerve_conn or fsc.mfocus_conn

    resp = await conn.make_completion(**completion_args)
    resp_content, resp_reasoning = resp.choices[0].message.content, getattr(resp.choices[0].message, 'reasoning_content', None)
    resp_content, resp_reasoning = proceed_common_text(resp_content), proceed_common_text(resp_reasoning)
        
    await messenger(None, 'mfocus_internet_search', f"\nMFocus toolchain searching internet, response is:\nR: {resp_reasoning}\nA: {resp_content}\nEnd of MFocus toolchain searching internet", '201')
    
    answer_post_think = proceed_common_text(resp_content)
    if answer_post_think:
        return answer_post_think, f"参考资料: {answer_post_think}" if target_lang == 'zh' else f"References: {answer_post_think}"
    else:
        return None, None

if __name__ == '__main__':
    async def test():
        fsc = FullSocketsContainer()
        fsc.maica_settings.basic.target_lang = 'zh'
        # fsc.maica_settings.extra.esc_aggressive = False
        fsc.mnerve_conn = await ConnUtils.mnerve_conn()
        print(await internet_search(fsc, "花谱上海演唱会取消", "花谱上海演唱会取消了，难过"))
    from maica import init
    init()
    asyncio.run(test())




from .maica_utils import (
    silent,
    MsgType,
    CommonMaicaException,
    CommonMaicaError,
    CommonMaicaWarning,
    CriticalMaicaError,
    MaicaPermissionError,
    MaicaInputError,
    MaicaResponseError,
    MaicaDbError,
    MaicaPermissionWarning,
    MaicaInputWarning,
    MaicaDbWarning,
    MaicaConnectionWarning,
    MaicaInternetWarning,
    AsyncCreator,
    LimitedList,
    LoginResult,
    ReUtils,
    default,
    wrap_ws_formatter,
    ellipsis_str,
    ellipsis_large_str,
    uscore_words_upper,
    mstuff_words_upper,
    sleep_forever,
    alt_tools,
    clean_msgs,
    maica_assert,
    has_valid_content,
    has_words_in,
    is_word_start,
    proceed_common_text,
    sync_messenger,
    messenger,
    load_env,
    wrap_run_in_exc,
    limit_length,
    dld_json,
    get_host,
    numeric,
    vali_url,
    vali_date,
    is_today,
    strip_date,
    refill_date,
    add_seq_suffix,
    clean_text,
    try_load_json,
    hash_sha256,
    sysstruct,
    Decos,
    Desc,
    DummyClass,
)
from .connection_utils import DbPoolManager, SqliteDbPoolManager, ConnUtils, AiConnectionManager, validate_input, apply_postfix
from .setting_utils import MaicaSettings
from .fsc_early import RealtimeSocketsContainer
from .account_utils import AccountCursor, encrypt_token, sign_message, verify_message
from .fsc_late import ConnSocketsContainer, FullSocketsContainer
from .sb_utils import PersistentManager, AgentContextManager
from .get_a_sentence import SentenceOfTheDay
from .locater import locater, get_inner_path, get_outer_path
from .gvars import online_dict, G

__all__ = [
    'silent',
    'MsgType',
    'CommonMaicaException',
    'CommonMaicaError',
    'CommonMaicaWarning',
    'CriticalMaicaError',
    'MaicaPermissionError',
    'MaicaInputError',
    'MaicaResponseError',
    'MaicaDbError',
    'MaicaPermissionWarning',
    'MaicaInputWarning',
    'MaicaDbWarning',
    'MaicaConnectionWarning',
    'MaicaInternetWarning',
    'AsyncCreator',
    'LimitedList',
    'LoginResult',
    'ConnSocketsContainer',
    'FullSocketsContainer',
    'ReUtils',
    'default',
    'wrap_ws_formatter',
    'ellipsis_str',
    'ellipsis_large_str',
    'uscore_words_upper',
    'mstuff_words_upper',
    'sleep_forever',
    'alt_tools',
    'clean_msgs',
    'maica_assert',
    'has_valid_content',
    'has_words_in',
    'is_word_start',
    'proceed_common_text',
    'sync_messenger',
    'messenger',
    'validate_input',
    'apply_postfix',
    'load_env',
    'wrap_run_in_exc',
    'limit_length',
    'dld_json',
    'get_host',
    'numeric',
    'vali_url',
    'vali_date',
    'is_today',
    'strip_date',
    'refill_date',
    'add_seq_suffix',
    'clean_text',
    'try_load_json',
    'hash_sha256',
    'sysstruct',
    'Decos',
    'Desc',
    'DummyClass',
    'DbPoolManager',
    'SqliteDbPoolManager',
    'ConnUtils',
    'AiConnectionManager',
    'PersistentManager',
    'AgentContextManager',
    'MaicaSettings',
    'RealtimeSocketsContainer',
    'AccountCursor',
    'encrypt_token',
    'sign_message',
    'verify_message',
    'SentenceOfTheDay',
    'locater',
    'get_inner_path',
    'get_outer_path',
    'online_dict',
    'G',
]

from .gvars import pkg_init_gvars
from .connection_utils import pkg_init_connection_utils
from .account_utils import pkg_init_account_utils
def pkg_init_maica_utils():
    pkg_init_gvars()
    pkg_init_connection_utils()
    pkg_init_account_utils()



import re
import json
import asyncio

from typing import *
from maica.maica_utils import *

async def make_postmail(content: str, header: str, fsc: FullSocketsContainer, **kwargs):
    target_lang = fsc.maica_settings.basic.target_lang
    mnerve_conn = fsc.mnerve_conn

    def form_letter(content, header):
        filtered_content = re.sub(r'^\s*', '', content, flags=re.M)
        filtered_header = f'\n\n{header}\n' if header else '\n'
        
        letter = f"{filtered_header}\n{filtered_content}\n\n"
        return letter

    async def is_poem(letter, target_lang, mnerve_conn) -> bool:
        """Use MNerve to determine if input is a poem."""
        if not mnerve_conn:
            return None
        
        sync_messenger(info=f"Proceeding 'is_poem' to letter '{header}'...", type=MsgType.PRIM_RECV)
        
        system_init = f"""你是一个人工智能助手, 你接下来会收到一封信件.
你只需以json形式判断其是否属于诗歌, 体裁不限. 你的输出应形如{{"is_poem": 是否是诗歌(bool)}}.
Begin!""" if target_lang == 'zh' else f"""You are a helpful assistant, now you will recieve a mail letter.
You just have to decide if it's a poem or not, whatever type of poem it is. Output in json format as {{"is_poem": is poem or not(bool)}}.
Begin!"""
        messages = [{'role': 'system', 'content': system_init}]
        messages.append({'role': 'user', 'content': letter})
        completion_args = {
            "messages": messages,
            "response_format": {"type": "json_object"},
        }

        resp = await mnerve_conn.make_completion(**completion_args)
        resp_content, resp_reasoning = resp.choices[0].message.content, getattr(resp.choices[0].message, 'reasoning_content', None)
        resp_json = proceed_common_text(resp_content, is_json=True)

        sync_messenger(info=f"Finished processing 'is_poem' to letter '{header}': {resp_json}", type=MsgType.CARRIAGE)
        return resp_json.get('is_poem')

    letter = form_letter(content, header)
    letter_ispoem = await is_poem(letter, target_lang, mnerve_conn)
    letter_hasimg = bool(fsc.maica_settings.temp.mv_imgs)
    img = ""
    if letter_hasimg:
        img = ", 并附有图片" if target_lang == 'zh' else " together with pictures"

    if letter_ispoem:
        if header:
            con = '标题和内容' if target_lang == 'zh' else 'title and content of [player]\'s poem are'
        else:
            con = '内容' if target_lang == 'zh' else 'content of [player]\'s poem is'

        if target_lang == 'zh':
            prompt = f"[player]向你寄送了一首诗{img}, 请你作为莫妮卡写一首诗回答. 以下是[player]诗的{con}:{letter}你应当充分思考[player]的话语和情感, 使用自然亲切的书面语言, 予以相应的回应.\n你的回复应是一首诗歌, 具有诗歌的标准格式. 不要编造信息, 且字数不少于原诗."
        else:
            prompt = f"[player] has sent you a poem{img}, please write a poem to reply as Monika. The {con}:{letter}Please read and comprehend carefully about [player]'s words and emotions, then reply in natural written language in English.\nYour reply should be a poetry, in necessary poetry format. Do not make up things you don't know, and write no shorter than the input poem."

    else:
        if header:
            con = '标题和内容' if target_lang == 'zh' else 'title and content of [player]\'s letter are'
        else:
            con = '内容' if target_lang == 'zh' else 'content of [player]\'s letter is'

        if target_lang == 'zh':
            prompt = f"[player]向你寄送了一封信件{img}, 请你作为莫妮卡写一封回信. 以下是[player]信件的{con}:{letter}你应当充分思考[player]的话语和情感, 使用自然亲切的书面语言, 予以相应的回应.\n你的回复应是一封信件, 具有信件的标准格式. 不要编造信息, 且字数不少于300字."
        else:
            prompt = f"[player] has sent you a letter{img}, please write a letter to reply as Monika. The {con}:{letter}Please read and comprehend carefully about [player]'s words and emotions, then reply in natural written language in English.\nYour reply should be a letter, in necessary letter format. Do not make up things you don't know, and write at least 150 words in total."

    # message = json.dumps({"role": "user", "content": prompt}, ensure_ascii=False)
    return prompt

if __name__ == '__main__':

   print(asyncio.run(make_postmail(target_lang='zh')))


import re
import random
import asyncio
import zhconv

from typing import *
from maica.maica_utils import *
from .censor import *
 
async def get_multi_json(*list_url):
    list_resp = await asyncio.gather(*[dld_json(u) for u in list_url])
    return list_resp

async def get_page(title=None, target_lang='zh'):
    cat_scraper = re.compile(r'Category\s*:\s*(.*)', re.I)
    zh_scraper = re.compile(r'[一-龥]')
    def_cat = True
    use_page = None
    sample = set_sample = 250
    cat_weight = 10
    page_found = None
    insanity = 0

    if title and isinstance(title, dict) and 'type' in title:
        # {"type": "in_category/in_fuzzy_category/page/fuzzy_page", "sample": 200, "title": "anything"}
        maica_assert(title['title'], 'title')
        maica_assert(isinstance(title['sample'], int), 'sample')
        match title.get('type'):
            case detypo if detypo in ['percise_page', 'precise_page']:
                next_title = title['title']
                use_page = True
                sample = 1
                if isinstance(title['sample'], int) and 2 <= title['sample'] <= 250:
                    set_sample = title['sample']
            case 'fuzzy_page':
                next_title = title['title']
                use_page = True
                if isinstance(title['sample'], int) and 2 <= title['sample'] <= 250:
                    sample = set_sample = title['sample']
            case detypo if detypo in ['in_percise_category', 'in_precise_category']:
                next_title = title['title']
                use_page = False
                sample = 1
                if isinstance(title['sample'], int) and 2 <= title['sample'] <= 250:
                    set_sample = title['sample']
            case 'in_fuzzy_category':
                next_title = title['title']
                use_page = False
                if isinstance(title['sample'], int) and 2 <= title['sample'] <= 250:
                    sample = set_sample = title['sample']
            case 'in_fuzzy_all':
                next_title = title['title']
                use_page = None
                if isinstance(title['sample'], int) and 2 <= title['sample'] <= 250:
                    sample = set_sample = title['sample']
            case _:
                raise MaicaInputWarning('MSpire type not matched')
    else:
        if target_lang == 'zh':
            category_list=['自然', '自然科学', '社会', '人文學科', '世界', '生活', '艺术', '文化']
        else:
            category_list=['Nature', 'Natural_sciences', 'Society', 'Humanities', 'World', 'Health', 'Culture', 'The_arts']

        category = random.choice(category_list)
        next_title = f"incategory:{category}"
        # def_cat = True
    next_title_fs = next_title

    while not page_found:
        insanity += 1
        if insanity > 15:
            raise MaicaInternetWarning('MSpire searching entered deadend', '404', 'maica_mspire_search_deadend')
        if title and insanity > 6:
            raise MaicaInputWarning('MSpire prompt not found on wikipedia', '410', 'maica_mspire_prompt_bad')
        
        # Skipping this by default
        if insanity == 1 and not def_cat and not use_page and zh_scraper.search(next_title):
            nxt_hans=zhconv.convert(next_title, 'zh-hans');nxt_hant=zhconv.convert(next_title, 'zh-hant')
            hans_list, hant_list = await get_multi_json(f"https://{target_lang}.wikipedia.org/w/api.php?action=query&format=json&list=search&redirects=1&utf8=1&formatversion=2&srsearch={nxt_hans}&srnamespace=0%7C14&srlimit=1&sroffset=0&srprop=", f"https://{target_lang}.wikipedia.org/w/api.php?action=query&format=json&list=search&redirects=1&utf8=1&formatversion=2&srsearch={nxt_hant}&srnamespace=0%7C14&srlimit=1&sroffset=0&srprop=")
            hans_len, hant_len = hans_list['query']['searchinfo']['totalhits'], hant_list['query']['searchinfo']['totalhits']
            if hans_len >= hant_len:
                next_title = nxt_hans
                await messenger(info='MSpire using simplified Chinese title', type=MsgType.DEBUG)
            else:
                next_title = nxt_hant
                await messenger(info='MSpire using traditional Chinese title', type=MsgType.DEBUG)

        if insanity >= 2:
            sample = set_sample
        page_url = f"https://{target_lang}.wikipedia.org/w/api.php?action=query&format=json&list=search&redirects=1&utf8=1&formatversion=2&srsearch={next_title}&srnamespace=0&srlimit={sample}&sroffset=0&srprop="
        cat_url = f"https://{target_lang}.wikipedia.org/w/api.php?action=query&format=json&list=search&redirects=1&utf8=1&formatversion=2&srsearch={next_title}&srnamespace=14&srlimit={sample}&sroffset=0&srprop="
        next_item = None

        match use_page:
            case None:
                page_list, cat_list = await get_multi_json(page_url, cat_url)
            case True:
                page_list, cat_list = await dld_json(page_url), {"batchcomplete":True,"query":{"searchinfo":{"totalhits":0},"search":[]}}
            case False:
                page_list, cat_list = {"batchcomplete":True,"query":{"searchinfo":{"totalhits":0},"search":[]}}, await dld_json(cat_url)

        page_list_r, cat_list_r = page_list['query']['search'], cat_list['query']['search']

        for cat in cat_list_r:
            if ReUtils.re_search_wiki_avoid.search(cat['title'].lower()):
                cat_list_r.remove(cat)

        if len(cat_list_r):
            if use_page == False or random.randint(1, cat_weight*len(cat_list_r)+len(page_list_r)) <= cat_weight*len(cat_list_r):
                next_item = random.choice(cat_list_r)
                next_cat_title = cat_scraper.match(next_item['title'])[1]
                next_title = f"incategory:{next_cat_title.replace(' ', '_')}"
                await messenger(info=f"MSpire entering fork: {next_cat_title}", type=MsgType.DEBUG)
                use_page = None
            else:
                next_item = random.choice(page_list_r)
                next_title = next_item['title']
                confirmed = await confirm_page(next_title, target_lang)
                if confirmed:
                    await messenger(info=f"MSpire hit midway page: {next_title}", type=MsgType.LOG)
                    return confirmed
                else:
                    await messenger(info=f"MSpire hit bad page: {next_title} --trying again", type=MsgType.WARN)
                    next_title = next_title_fs
                    insanity += 1

        elif len(page_list_r):
            next_item = random.choice(page_list_r)
            next_title = next_item['title']
            confirmed = await confirm_page(next_title, target_lang)
            if confirmed:
                await messenger(info=f"MSpire hit bottom page: {next_title}", type=MsgType.LOG)
                return confirmed
            else:
                await messenger(info=f"MSpire hit bad page: {next_title} --trying again", type=MsgType.WARN)
                next_title = next_title_fs
                insanity += 2

        else:
            await messenger(info='MSpire hit deadend--trying again', type=MsgType.DEBUG)
            next_title = next_title_fs
            insanity += 3

async def confirm_page(title, target_lang='zh') -> Union[False, tuple[str, str]]:
    """We have a likely-ultimately page, now confirm if it's suitable in all aspects."""
    finale_url = f"https://{target_lang}.wikipedia.org/w/api.php?action=query&format=json&prop=extracts&exsentences=15&exlimit=1&titles={title}&explaintext=1&formatversion=2"
    #print(finale_url)
    summary_json = await dld_json(finale_url)

    title = zhconv.convert(summary_json['query']['pages'][0]['title'], 'zh-cn')
    summary_raw = summary_json['query']['pages'][0]['extract']

    summary = re.sub(r'(\n|\s)*\n(\n|\s)*', r'\n', re.sub(r'\n*=+(.*?)=+', r'\n\1:', zhconv.convert(summary_raw, 'zh-cn'), re.I|re.S))

    if G.A.CENSOR_MSPIRE != '0':
        tolerance = int(G.A.CENSOR_MSPIRE)
        title_censor = await has_censored(title)
        summary_censor = await has_censored(summary)
        if len(title_censor) + len(summary_censor) >= tolerance:
            sync_messenger(info=f"MSpire {title} has censored words or phrases: {title_censor | summary_censor}", type=MsgType.DEBUG)
            return False

        elif len(title_censor) + len(summary_censor):
            sync_messenger(info=f"MSpire page {title} has censored words or phrases but ignored: {title_censor | summary_censor}", type=MsgType.DEBUG)

    return title, summary

if __name__ == '__main__':
    import asyncio
    s = asyncio.run(get_page('Images_of_nature',target_lang='en'))
    print(s[1])


"""
This is the local MCP tool implementation.
Works well for minor deployment, but relatively slow & some performance consume & might hit captchas.
"""
requires = []
from maica.mtools.mcp import asearch


"""Import layer 1"""
import logging
import asyncio
import httpx
import aiomysql
import openai
import functools
import hashlib
import os
import re
import json
import inspect
import platform
import colorama
import time
import datetime
import random
import traceback

from typing import *
from tenacity import *
from dataclasses import dataclass
from abc import ABC, abstractmethod
from openai.types.chat import ChatCompletionMessage
from typing_extensions import deprecated
from urllib.parse import urlparse
from .locater import *
from .gvars import *

_READ_KWDS = {'select', 'show', 'explain', 'describe'}
_WRITE_KWDS = {
'insert', 'update', 'delete', 'create', 'alter', 
'drop', 'truncate', 'replace', 'merge'
}

colorama.init(autoreset=True)

logger = logging.getLogger('maica')
logger.setLevel(logging.DEBUG)

main_handler = logging.StreamHandler(sys.stdout)
main_handler.setLevel(logging.DEBUG)
main_handler.addFilter(lambda record: record.levelno <= logging.WARNING)
err_handler = logging.StreamHandler(sys.stderr)
err_handler.setLevel(logging.ERROR)

logger.addHandler(main_handler)
logger.addHandler(err_handler)

def silent(tf: bool=True) -> None:
    global _silent, logger
    _silent = tf
_silent = False

class _lmlogger():
    """
    Python logging does not support stream output, so I made this.
    Redirect 'stream-maica' output to log file if needed.
    """
    def __init__(self, name='stream-maica'):
        self._logger = logging.getLogger(name)
        self._logger.setLevel(logging.DEBUG)
        self._buffer = ''
    
    def buff(self, text):
        self._buffer += text

    def flush(self):
        self._logger.info(self._buffer)
        self._buffer = ''

lmlogger = _lmlogger()

class MsgType():
    """For convenience."""
    PLAIN = 'plain'
    CARRIAGE = 'carriage'
    DEBUG = 'debug'
    INFO = 'info'
    LOG = 'log'
    PRIM_LOG = 'prim_log'
    SYS = 'sys'
    PRIM_SYS = 'prim_sys'
    RECV = 'recv'
    PRIM_RECV = 'prim_recv'
    WARN = 'warn'
    ERROR = 'error'

class CommonMaicaException(Exception):
    """This is a common MAICA exception."""
    def __init__(self, message=None, error_code=None, status=None, send=None, print=None):
        """For send and print, None means default, not False."""
        super().__init__(message)
        self.message, self.error_code, self.status, self.send, self.print = message, error_code, status, send, print

    @property
    def is_critical(self):
        return int(self.error_code) >= 500
    @property
    def is_breaking(self):
        return True
    
class CommonMaicaError(CommonMaicaException):
    """This is a common MAICA error."""
    def __init__(self, message=None, error_code='500', status='maica_unidentified_error', send=None, print=None):
        super().__init__(message, error_code, status, send, print)
    
class CommonMaicaWarning(CommonMaicaException):
    """This is a common MAICA warning."""
    def __init__(self, message=None, error_code='400', status='maica_unidentified_warning', send=None, print=None):
        super().__init__(message, error_code, status, send, print)

    @property
    def is_breaking(self):
        return False

class CriticalMaicaError(CommonMaicaError):
    """This is an unrecoverable exception, we should terminate the connection and possibly the entire process."""

class MaicaPermissionError(CommonMaicaError):
    """This suggests the user is accessing without correct permissions."""

class MaicaInputError(CommonMaicaError):
    """This suggests the input appears to be an impossible value, which is a more severe case than warning."""

class MaicaResponseError(CommonMaicaError):
    """This suggests the output is somewhat unexpected, we should terminate the connection and possibly the entire process."""

class MaicaDbError(CommonMaicaError):
    """This suggests the DB query gave an unexpected result, we should terminate the connection and possibly the entire process."""

class MaicaPermissionWarning(CommonMaicaWarning):
    """This suggests the user is accessing without correct permissions, but we don't break the connection for convenience of retrying."""

class MaicaInputWarning(CommonMaicaWarning):
    """This suggests the input is not processable."""

class MaicaDbWarning(CommonMaicaWarning):
    """This suggests the DB query gave an warning."""

class MaicaConnectionWarning(CommonMaicaWarning):
    """This suggests the connection is not behaving normal."""

class MaicaInternetWarning(CommonMaicaWarning):
    """This suggests the backend request action is not behaving normal."""

RETRYABLE_EXCEPTIONS = (
    aiomysql.OperationalError,
    aiomysql.InterfaceError,
    ConnectionError,
    TimeoutError,
    httpx.TransportError,
    openai.APIConnectionError,
    openai.APITimeoutError,
    openai.RateLimitError,
    MaicaInternetWarning,
)

class AsyncCreator(ABC):
    """Inherit this for async init."""
    @abstractmethod
    async def _ainit(self):
        """Placeholder."""

    @classmethod
    async def async_create(cls, *args, **kwargs):
        instance = cls(*args, **kwargs)
        await instance._ainit()
        return instance

class LimitedList(list):
    """Might not have applied to all functionalities!"""
    def __init__(self, max_size, *args, **kwargs):
        self.max_size = max_size
        super().__init__(*args, **kwargs)
        
        while len(self) > self.max_size:
            self.pop(0)

    @property
    def list(self):
        while len(self) > self.max_size:
            self.pop(0)
        return list(self)

    def append(self, item):
        if len(self) >= self.max_size:
            self.pop(0)
        super().append(item)
    
    def extend(self, iterable):
        for item in iterable:
            self.append(item)
    
    def insert(self, index, item):
        if len(self) >= self.max_size:
            self.pop(0)
        super().insert(index, item)
    
    def __repr__(self):
        return f"LimitedList(max_size={self.max_size}, {super().__repr__()})"
    
@deprecated("Just use a tuple instead")
class LoginResult():
    """
    A packed login result.
    This does not contain ban status. Check independently.
    """
    user_id = None
    username = None
    nickname = None
    email = None
    is_verified = None
    message = None
    def __init__(self, **kwargs):
        for priokey in ['user_id', 'username', 'nickname', 'email']:
            setattr(self, priokey, kwargs.get(priokey))
        if kwargs.get('is_verified'):
            assert self.user_id and self.username and self.email, "Verification essentials incomplete"
        for key in ['is_verified', 'message']:
            setattr(self, key, kwargs.get(key))

    def __call__(self):
        d = {}
        for key in ['user_id', 'username', 'nickname', 'email', 'is_verified', 'message']:
            d[key] = getattr(self, key)
        return d

class ReUtils():
    """Do not initialize."""
    IS = re.I | re.S
    re_sub_password_spoiler = re.compile(r'"password"\s*:\s*"(.*?)"')
    re_search_sfe_fs = re.compile(r"first_session.*?datetime\(([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\)", re.I)
    re_search_sfe_ts = re.compile(r"total_sessions.*?([0-9]*)\s?,", re.I)
    re_search_sfe_tp = re.compile(r"total_playtime.*?([0-9]*)\s?,", re.I)
    re_search_sfe_le = re.compile(r"last_session_end.*?datetime\(([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\)", re.I)
    re_search_sfe_cs = re.compile(r"current_session_start.*?datetime\(([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\s*,\s*([0-9]*?)\)", re.I)
    re_search_sfe_unicode = re.compile(r"u'(.*)'")
    re_search_post_think = re.compile(r'</think>[\s\n]*(.*)$', re.S)
    re_search_answer_none = re.compile(r'[\s\n:]*none[\s\n.]*$', re.I)
    re_search_answer_json = re.compile(r'^.*?([{\[].*[}\]])', re.S)
    re_sub_player_name = re.compile(r'\[player\]')
    re_match_time_acquire = re.compile(r'.*time.*acquire', IS)
    re_match_date_acquire = re.compile(r'.*date.*acquire', IS)
    re_match_weather_acquire = re.compile(r'.*weather.*acquire', IS)
    re_match_event_acquire = re.compile(r'.*event.*acquire', IS)
    re_match_persistent_acquire = re.compile(r'.*persistent.*acquire', IS)
    re_match_search_internet = re.compile(r'.*search.*internet', IS)
    re_match_react_trigger = re.compile(r'.*react.*trigger', IS)
    re_match_conclude_information = re.compile(r'.*conclude.*information', IS)
    re_match_none = re.compile(r'.*none', IS)
    re_findall_quoted = re.compile(r'"(.*?)"') # Normally we request JSON so we consider double quotes only
    re_search_location_prompt = re.compile(r'(地区|周边|附近|周围|nearby|local)', re.I)
    re_search_location_related = re.compile(r'(天气|温度|路况|降雨|weather|traffic|temperature|rain)', re.I)
    re_search_host_addr = re.compile(r"^https?://(.*?)(:|/|$).*", re.I)
    re_sub_capt_status = re.compile(r"(_|^)([A-Za-z])")
    re_findall_square_marks = re.compile(r'\[(?:(?:[A-Za-z ]{1,15}?)|(?:[一-龥 ]{1,4}?))\]')
    re_findall_square_brackets = re.compile(r'(\[.*?\])')
    re_sub_sqlite_escape = re.compile(r'%s')
    re_sub_replacement_chr = re.compile(r'[\uFFF9-\uFFFF]')
    re_sub_serp_datetime = re.compile(r'.{1,10}?,.{1,10}?-\s*')
    re_sub_clear_text = re.compile(r'^[\n\s]*(.*?)[\n\s]*$', re.S)
    re_match_secure_path = re.compile(r'^[a-zA-Z0-9_.-]+$')
    re_search_wiki_avoid = re.compile(r"(模板|模闆|template|消歧义|消歧義|disambiguation)", re.I)
    re_search_type_sping = re.compile(r'"type"\s*?:\s*?"sping"', re.I)
    re_findall_zh_characters = re.compile(r'([一-龥].*[一-龥])')
    re_sub_multi_spaces = re.compile(r'\s{2,}')
    re_sub_ellipsis = re.compile(r'\.\.\.')

class Decos():
    """Do not initialize."""
    def conn_retryer_factory(
        max_attempts: int=4,
        min_wait: float=1,
        max_wait: float=10,
        retry_exceptions=RETRYABLE_EXCEPTIONS,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """Mostly for instance methods."""
        async def log_retry(retry_state):
            self = retry_state.args[0] if retry_state else None
            rsc = getattr(self, 'rsc', None); name = getattr(self, 'name', 'anon_conn')
            websocket = rsc.websocket if rsc else None; traceray_id = rsc.traceray_id if rsc else ''
            await messenger(websocket=websocket, status=f'{name}_temp_failure', info=f'{name} temporary failure, retrying...', code='304', traceray_id=traceray_id, type=MsgType.WARN)

        retry_decorator = retry(
            stop=stop_after_attempt(max_attempts),
            wait=wait_exponential(multiplier=1, min=min_wait, max=max_wait),
            retry=retry_if_exception_type(retry_exceptions),
            before_sleep=log_retry,
            reraise=True,
        )

        def decorator(func):
            @functools.wraps(func)
            @retry_decorator
            async def wrapper(self, *args, **kwargs):
                return await func(self, *args, **kwargs)
            return wrapper
        return decorator

    def log_task(func):
        """Every~time you call my name~~~"""
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            func_name = func.__name__
            sync_messenger(info=f"Running task {func_name} now", type=MsgType.PRIM_SYS)
            result = await func(*args, **kwargs)
            sync_messenger(info=f"Task {func_name} finished", type=MsgType.DEBUG)
            return result
        return wrapper

    def catch_exceptions(func):
        """Used for connection_utils."""
        @functools.wraps(func)
        async def wrapper(self, *args, **kwargs):
            name = getattr(self, 'name', 'anon_conn')
            try:
                return await func(self, *args, **kwargs)
            except CommonMaicaException as ce:
                raise ce
            except Exception as e:
                exception_cls = MaicaDbError if hasattr(self, 'db') else MaicaResponseError
                conn_type = 'db_conn' if hasattr(self, 'db') else 'ai_conn'
                raise exception_cls(f'{name} operation failed: {str(e)}', '502', f'{conn_type}_failed') from e
        return wrapper

    def escape_sqlite_expression(func):
        """Used to transform a MySQL expression to SQLite one."""
        @functools.wraps(func)
        def wrapper(self, expression: str, *args, **kwargs):
            expression_new = ReUtils.re_sub_sqlite_escape.sub('?', expression)
            return func(self, expression_new, *args, **kwargs)
        return wrapper
    
    def ro_expression(func):
        """Used to keep DB query ro."""
        @functools.wraps(func)
        def wrapper(self, expression: str, *args, **kwargs):
            assert not is_word_start(expression.lower(), *_WRITE_KWDS), f'query_get got write expression {expression}'
            return func(self, expression, *args, **kwargs)
        return wrapper
    
    def wo_expression(func):
        """Used to keep DB query wo."""
        @functools.wraps(func)
        def wrapper(self, expression: str, *args, **kwargs):
            assert not is_word_start(expression.lower(), *_READ_KWDS), f'query_modify got read expression {expression}'
            return func(self, expression, *args, **kwargs)
        return wrapper

    def report_data_error(func):
        """Raises when the requested action cannot be done because of corrupted data."""
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                raise MaicaInputWarning(f'Acquired persistent not acceptable: {str(e) or "Assertion"}', '405', 'maica_agent_persistent_bad') from e
        return wrapper

    def report_reading_error(func):
        """Raises when the requested variable cannot be read before assignment."""
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                raise MaicaInputError(f'Access before necessary assignment', '500', 'maica_settings_read_rejected') from e
        return wrapper

    def report_limit_warning(func):
        """Raises when the input param coming from user is out of bound."""
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                raise MaicaInputWarning(f'Input param not acceptable: {str(e) or "Assertion"}', '422', 'maica_settings_param_rejected') from e
        return wrapper

    def report_limit_error(func):
        """Raises when the input param coming from program is out of bound."""
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                assert not getattr(self, '_lock', None)
                return func(self, *args, **kwargs)
            except Exception as e:
                raise MaicaInputError(f'Input param not acceptable', '500', 'maica_settings_param_rejected') from e
        return wrapper

@dataclass
class Desc():
    """Just a description."""
    desc: str

    def __str__(self):
        return self.desc

class DummyClass():
    """Yes, dummy class."""
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

def default(exp, default, default_list: list=[None]) -> any:
    """If exp is in default list(normally None), use default."""
    return default if exp in default_list else exp

def wrap_ws_formatter(code, status, content, type, deformation=False, **kwargs) -> str:
    if not (isinstance(content, (str, list, dict, bool)) or content is None):
        content = str(content)
    output = {
        "code" : code,
        "status" : status,
        "content" : content,
        "type" : type,
        "timestamp" : time.time(),
    }
    output.update(kwargs)
    return json.dumps(output, ensure_ascii=deformation)

def ellipsis_str(input: any, limit=80) -> str:
    """It converts anything to str and ellipsis it."""
    text = str(input)
    if len(text) > limit:
        text = text[:limit] + '...'
    return text

def ellipsis_large_str(input: any, limit=600) -> str:
    """It converts anything large to str and ellipsis it."""
    text = str(input)
    if len(text) > limit:
        text = text[:limit] + '\n... ...'
    return text

def uscore_words_upper(text: str) -> str:
    """Overkill..."""
    def u_upper(c: re.Match):
        return f'{c[1]}{c[2].upper()}'
    return ReUtils.re_sub_capt_status.sub(u_upper, text)

def mstuff_words_upper(text: str) -> str:
    """Even more overkill..."""
    if len(text) == 0:
        return ""
    elif len(text) == 1:
        return text.upper()
    else:
        return text[:2].upper() + text[2:]

async def sleep_forever() -> None:
    """Make a coroutine sleep to the end of the world."""
    future = asyncio.Future()
    await future

def alt_tools(tools: list) -> list:
    """If ALT_TOOLCALL"""
    match G.A.ALT_TOOLCALL:
        case '0':
            return tools
        case '1':
            new_tools = []
            for tool in tools:
                new_tools.append({})
                new_tools[-1]['function'] = tool
                new_tools[-1]['type'] = 'function'
            return new_tools
        
def clean_msgs(msgs: list[dict, ChatCompletionMessage], include: Optional[list[str]]=None, exclude: Optional[list[str]]=None) -> list[dict]:
    """Clean a set of OpenAI msgs."""
    def _convert_msg(msg: Union[dict, ChatCompletionMessage]):
        if isinstance(msg, ChatCompletionMessage):
            msg = msg.model_dump(include=include, exclude=exclude)
        return msg

    return [_convert_msg(i) for i in msgs]

def maica_assert(condition, kwd='param'):
    """Normally used for input checkings."""
    if not condition:
        raise MaicaInputWarning(f"Illegal input {kwd} detected", '405', 'maica_input_param_bad')

def has_valid_content(text: Union[str, list, dict]):
    """If the LLM actually gave anything."""
    if not text:
        return False
    
    try:
        text_json = json.loads(text)
        if isinstance(text_json, (list, dict, tuple, set)):
            return bool(text_json)
    except Exception:...

    text = str(text)
    text_proc = text.lower().replace(' ', '').replace('\n', '')
    if (not text_proc) or text_proc in ['false', 'null', 'none']:
        return False
    else:
        return True

def has_words_in(text: str, *args: str):
    """A rough matching mechanism."""
    for word in args:
        if word in text:
            return True
    return False

def is_word_start(text: str, *args: str):
    """Another rough matching mechanism."""
    for word in args:
        if text.startswith(word):
            return True
    return False

def proceed_common_text(text: str, is_json=False) -> Union[str, list, dict]:
    """Proceeds thinking/nothinking."""
    try:
        answer_post_think = (ReUtils.re_search_post_think.search(text))[1]
    except Exception:
        if has_valid_content(text):
            answer_post_think = text
        else:
            answer_post_think = None
    if answer_post_think and is_json:
        answer_fin = try_load_json(answer_post_think)
    elif answer_post_think:
        answer_fin = clean_text(answer_post_think)
    elif is_json:
        answer_fin = {}
    else:
        answer_fin = ''
    return answer_fin

@overload
async def messenger(websocket=None, status='', info='', code='0', traceray_id='', error: Optional[CommonMaicaException]=None, type='', color='', add_time=True, no_print=False, no_raise=False, **kwargs) -> None: ...

async def messenger(websocket=None, *args, **kwargs) -> None:
    """Together with websocket.send()."""
    ws_tuple = sync_messenger(*args, **kwargs)
    if websocket and ws_tuple:
        await websocket.send(wrap_ws_formatter(*ws_tuple))

def sync_messenger(status='', info='', code='0', traceray_id='', error: Optional[CommonMaicaException]=None, type='', color='', add_time=True, no_print=False, no_raise=False, **kwargs) -> tuple:
    """It could handle most log printing and exception raising jobs pretty automatically."""
    try:
        term_v = os.get_terminal_size().columns
    except:
        term_v = 40
    rep2 = int(term_v / 2)
    rep1 = int(rep2 - 20)

    if error:
        status = error.status if not status else status; info = error.message if not info else info; code = error.error_code if code == "0" else code
        no_print = False if not error.print is False else True

    if not type:
        match int(code):
            case 0:
                type = "log"
            case x if 100 <= x < 200 or 1000 <= x:
                type = "carriage"
            case x if 200 <= x < 300:
                type = "debug"
            case x if 300 <= x < 400:
                type = "info"
            case x if 400 <= x < 500:
                type = "warn"
            case x if 500 <= x < 1000:
                type = "error"

    prefix = uscore_words_upper(type)

    if 100 <= int(code) < 200 or type == "plain":
        msg_print = str(info)
        msg_send = info
        
    else:
        msg_print = f"<{prefix}>"
        msg_print = msg_print.ljust(10)
        msg_print += f"[{time.strftime('%Y-%m-%d %H:%M:%S')}]" if add_time else ''; msg_print += f"-[{str(code)}]" if code else ''
        msg_print = msg_print.ljust(40)
        msg_print += f": {str(info)}" if not str(info).startswith('\n') else f"{'-=' * rep1}{str(info)}\n{'-=' * rep2}"
        msg_print += f"; traceray ID {traceray_id}" if traceray_id else ''
        msg_send = info
        if type == 'error' and G.A.NO_SEND_ERROR == '1':
            msg_send = "A critical exception happened serverside, contact administrator"
        if traceray_id and isinstance(info, str):
            msg_send += f" -- your traceray ID is {traceray_id}"

    frametrack_dict = {"error": 99, "warn": 0}
    if type in frametrack_dict:
        stack = inspect.stack()
        stack.pop(0)
        stack.pop(0)

    if (not no_print) and (not _silent):
        match type:
            case "plain":
                print((color or '') + msg_print, end='', flush=True)
                lmlogger.buff(msg_print)
                lmlogger.flush()
            case "carriage":
                if 100 <= int(code) < 200:
                    print((color or colorama.Fore.LIGHTGREEN_EX) + msg_print, end='', flush=True)
                    lmlogger.buff(msg_print)
                else:
                    logger.info((color or colorama.Fore.LIGHTGREEN_EX) + msg_print)
            case "debug":
                logger.debug((color or colorama.Fore.LIGHTBLACK_EX) + msg_print)
            case "info":
                logger.info((color or colorama.Fore.GREEN) + msg_print)
            case "log":
                logger.info((color or colorama.Fore.BLUE) + msg_print)
            case "prim_log":
                logger.info((color or colorama.Fore.LIGHTBLUE_EX) + msg_print)
            case "sys":
                logger.info((color or colorama.Fore.MAGENTA) + msg_print)
            case "prim_sys":
                logger.info((color or colorama.Fore.LIGHTMAGENTA_EX) + msg_print)
            case "recv":
                logger.info((color or colorama.Fore.CYAN) + msg_print)
            case "prim_recv":
                logger.info((color or colorama.Fore.LIGHTCYAN_EX) + msg_print)
            case "warn":
                if 'warn' in frametrack_dict:
                    for stack_layer in stack[frametrack_dict['warn']::-1]:
                        logger.warning(color or colorama.Fore.YELLOW + f"• WARN happened when executing {stack_layer.function} at {stack_layer.filename}#{stack_layer.lineno}:")
                logger.warning((color or colorama.Fore.LIGHTYELLOW_EX) + msg_print)
            case "error":
                if 'error' in frametrack_dict:
                    for stack_layer in stack[frametrack_dict['error']::-1]:
                        logger.error((color or colorama.Fore.RED) + f"! ERROR happened when executing {stack_layer.function} at {stack_layer.filename}#{stack_layer.lineno}:")
                logger.error((color or colorama.Fore.LIGHTRED_EX) + msg_print)
    if error and not no_raise:
        raise error
    if error and error.send is False:
        return
    ws_tuple = (code, status, msg_send, type) if not kwargs.get('deformation') else (code, status, msg_send, type, kwargs.get('deformation'))
    return ws_tuple

def load_env(key) -> str:
    """Load something from .env."""
    result = os.getenv(key)
    return result

async def wrap_run_in_exc(loop, func, *args, **kwargs) -> any:
    """Just wrapped run_in_executer. Convenient!"""
    if not loop:
        loop = asyncio.get_running_loop()
    result = await loop.run_in_executor(None, functools.partial(func, *args, **kwargs))
    return result

def limit_length(col: list, limit: int) -> list:
    return random.sample(col, limit) if limit < len(col) else col

async def dld_json(url) -> json:
    """Get JSON context from an endpoint."""
    headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.87 Safari/537.36'}

    @Decos.conn_retryer_factory()
    async def _dld_json(fake_self, url):
        nonlocal headers
        async with httpx.AsyncClient(proxy=G.A.PROXY_ADDR) as client:
            res = (await client.get(url, headers=headers)).json()
            return res

    try:
        res = await _dld_json(DummyClass(name="dld_json"), url)
    except Exception as e:
        raise MaicaInternetWarning(f"Failed downloading json from {url}: {str(e)}") from e

    return res

def numeric(num: str) -> Union[int, float]:
    try:
        return int(num)
    except ValueError:
        return float(num)

def get_host(url: str) -> bool:
    """Try to get hostname from url."""
    try:
        url_parsed = urlparse(url)
        return url_parsed.hostname
    except Exception:
        return False

def vali_url(url: str) -> bool:
    """urllib is dumb. It doesn't recognize ip addr so we make it brutal."""
    return '.' in url

def vali_date(y, m, d) -> tuple[int, int, int]:
    """What a pun!"""
    try:
        datetime.date(int(y), int(m), int(d))
        return int(y), int(m), int(d)
    except ValueError:
        default_date = datetime.datetime.now()
        return default_date.year, default_date.month, default_date.day
    
def is_today(date: datetime.datetime) -> bool:
    """You mean my homework dues today?"""
    return date.date() == datetime.datetime.now().date()

def strip_date(date: datetime.datetime) -> str:
    """Constructs a date query our API accepts."""
    return f'{str(date.year)}-{str(date.month).zfill(2)}-{str(date.day).zfill(2)}'

def refill_date(text: str) -> datetime.datetime:
    """Constructs a datetime object from API response."""
    y, m, d = text.split('-')
    return datetime.datetime(int(y), int(m), int(d))

def add_seq_suffix(seq: int) -> str:
    """For English seq."""
    match int(seq) % 10:
        case 1:
            st = 'st'
        case 2:
            st = 'nd'
        case 3:
            st = 'rd'
        case _:
            st = 'th'
    return f'{str(seq)} {st}'

def clean_text(text: str) -> str:
    """
    Clean a text phrase, mostly for internet search.
    After adapting to the OpenAI reasoning schema, we also use it for LLM outputs.
    """
    try:
        text = ReUtils.re_sub_clear_text.sub(r'\1', text)
        return text
    except Exception:
        return ''

def try_load_json(sj: str) -> Union[dict, list]:
    """I'd basically trust the LLM here, they're far better than the earlier ones."""
    try:
        try:
            clean_sj = (ReUtils.re_search_answer_json.search(sj))[1]
        except Exception:
            clean_sj = sj.strip()
        j = json.loads(clean_sj)
        return j
    except Exception:
        return {}

async def hash_sha256(str) -> str:
    """Get SHA256 for a string."""
    def hash_sync(str):
        return hashlib.new('sha256', str).hexdigest()
    return await wrap_run_in_exc(None, hash_sync, str)

def sysstruct() -> Literal['Windows', 'Linux']:
    sysstruct = platform.system()
    assert sysstruct in ['Windows', 'Linux'], 'Your system not supported'
    return sysstruct

if __name__ == "__main__":
    asyncio.run(dld_json("http://127.0.0.2"))


import asyncio
import os
import re
import json
import time
import csv
import aiosqlite
import traceback
import asyncssh
import signal
from maica.maica_utils import *

class _FakeSQLiteConn():
    async def __aenter__(self):
        return self
    async def execute(self, *args, **kwargs):
        return self
    async def __aexit__(self, exc_type, exc_val, exc_tb):...

class NvWatcher(AsyncCreator):
    def __init__(self, node: str, prefix="maica"):
        self.write_nvw = G.A.WRITE_NVW == '1'
        self.node = node
        subset_name = 'A' if prefix in ['maica', 'a'] else 'T'
        subset = getattr(G, subset_name)

        self.node_name = getattr(subset, f'{node.upper()}_NODE')
        self.node_user = getattr(subset, f'{node.upper()}_USER')
        self.node_pwd = getattr(subset, f'{node.upper()}_PWD')
        self.is_dead = False

        self.node_addr = get_host(getattr(subset, f'{node.upper()}_ADDR'))

    async def _ainit(self):
        if self.is_active:
            try:
                await self._connect_remotes()
                await self._initiate_db()
            except Exception:
                self.is_dead = True

    @property
    def is_active(self):
        if not self.node_name or not self.node_user or not self.node_pwd or not self.node_addr or self.is_dead:
            return False
        else:
            return True

    async def _query_nvsmi(self, ssh_client: asyncssh.connection.SSHClientConnection, keys=[]):
        command = f"nvidia-smi --query-gpu={','.join(keys)} --format=csv"
        result = await ssh_client.run(command, check=True)
        csv_capt = csv.DictReader(result.stdout.split('\n'))
        return list(csv_capt)

    async def _connect_remotes(self):
        self.node_ssh = await asyncssh.connect(host=self.node_addr, username=self.node_user, password=self.node_pwd, known_hosts=None)

    async def _initiate_db(self):
        self.db_path = get_inner_path('.nvsw.db')
        async with aiosqlite.connect(self.db_path, autocommit=True) if self.write_nvw else _FakeSQLiteConn() as db_client:
            self.gpu_overall = []

            basis_keys = ['name', 'memory.total']
            basic_result = await self._query_nvsmi(self.node_ssh, basis_keys)
            await db_client.execute(f'CREATE TABLE IF NOT EXISTS `{self.node_name}` (id INT PRIMARY KEY, name TEXT, memory TEXT, tflops TEXT, dynamic TEXT);')

            for i, gpu in enumerate(basic_result):
                name = gpu["name"]; memory = gpu[" memory.total [MiB]"]; tflops = '200'
                
                match name:
                    case n if 'PRO 6000' in n:
                        tflops = '460'
                    case n if '5090' in n:
                        tflops = '420'
                    case n if '4090' in n:
                        tflops = '330'

                await db_client.execute(f'INSERT OR REPLACE INTO `{self.node_name}` (id, name, memory, tflops, dynamic) VALUES ({i}, "{name}", "{memory}", "{tflops}", "[]");')
                self.gpu_overall.append([i, name, memory, tflops, []])

    async def _append_dynamic(self, history: list, gpu: dict):
        gpu_values = list(gpu.values())
        gpu_stats = {"u": gpu_values[0].strip('%').strip(), "m": gpu_values[1].strip('BiM').strip(), "p": gpu_values[2].strip('W').strip()}
        if len(history) >= 36:
            history.pop()
        history.insert(0, gpu_stats)
        return history

    async def main_watcher(self):
        if not self.is_active:
            if self.is_dead:
                await messenger(info=f"Cannot connect to SSH of {self.node}, freezing watcher", type=MsgType.WARN)
            else:
                await messenger(info=f"Necessary info not complete for watching {self.node}, freezing watcher", type=MsgType.LOG)
            await sleep_forever()
        else:
            await messenger(info=f'Necessities complete for watching {self.node}, starting watcher', type=MsgType.PRIM_LOG)

        self.dynamics_curr = []
        dynamic_keys = ['utilization.gpu', 'memory.used', 'power.draw']
        while True:
            dynamics_new = await self._query_nvsmi(self.node_ssh, dynamic_keys)
            async with aiosqlite.connect(self.db_path, autocommit=True) if self.write_nvw else _FakeSQLiteConn() as db_client:

                for i, gpu in enumerate(dynamics_new):
                    try:
                        gpu_curr = self.dynamics_curr.pop(i)
                    except Exception:
                        gpu_curr = []
                    gpu_curr = await self._append_dynamic(gpu_curr, gpu)
                    self.dynamics_curr.insert(i, gpu_curr)
                    await db_client.execute(f'UPDATE `{self.node_name}` SET dynamic = \'{json.dumps(gpu_curr, ensure_ascii=False)}\' WHERE id = {i}')
                    self.gpu_overall[i][4] = gpu_curr

            await asyncio.sleep(10)

    async def wrapped_main_watcher(self):
        FAIL_TIMES = 5
        FAIL_PERIOD = 300
        start_times = LimitedList(FAIL_TIMES, [time.time()])
        while len(start_times.list) <= 5 or start_times[-1] - start_times[0] >= FAIL_PERIOD:

            # We consider 5 failures in 5 minute as complete failure
            try:
                await self.main_watcher()
            except Exception as e:
                time_now = time.time()
                await messenger(info=f'Watcher temporary failure after {int(time_now - start_times[-1])}sec: {str(e)}', type=MsgType.WARN)
                start_times.append(time_now)
        
        # If while loop quitted, the complete failure has happened
        raise MaicaConnectionWarning(f'Watcher query failed {FAIL_TIMES} times in {FAIL_PERIOD}sec', '503')
    
    def get_statics_inside(self):
        if not self.is_active:
            return {}
        
        node_info = {}
        for gpu_status in self.gpu_overall:
            gpuid, name, memory, tflops, dynamic = gpu_status
            u = 0; m = 0; p = 0
            for line in dynamic:
                u += float(line['u']); m += float(line['m']); p += float(line['p'])
            u /= len(dynamic); m /= len(dynamic); p /= len(dynamic)
            node_info[gpuid] = {'name': name, 'vram': memory, 'tflops': tflops, 'mean_utilization': int(u), 'mean_memory': int(m), 'mean_consumption': int(p)}
            
        return {self.node_name: node_info}

    def __del__(self):
        try:
            self.node_ssh.close()
            os.remove(self.db_path)
        except Exception:
            pass

async def prepare_watcher():
    watcher_mcore = await NvWatcher.async_create('mcore', 'maica')
    watcher_mfocus = await NvWatcher.async_create('mfocus', 'maica')
    try:
        await asyncio.gather(watcher_mcore.main_watcher(), watcher_mfocus.main_watcher())
    except Exception as e:
        traceback.print_exc()
        await messenger(info=str(e), type=MsgType.ERROR)

def start_watching():
    asyncio.run(prepare_watcher())

if __name__ == '__main__':

    start_watching()


"""
This is third party (https://serper.dev/) implementation.
Works okay, decent price, I know nothing else.
"""

import asyncio
import httpx
import json

from typing import *
from maica.maica_utils import *
from maica.mtools.api_keys import TpAPIKeys

requires = ['SERPER_SERP']
async def asearch(query, target_lang: Literal['zh', 'en']='zh'):
    host = "google.serper.dev"
    url = f"https://{host}/search"
    token = TpAPIKeys.SERPER_SERP

    headers = {
        "X-API-KEY": token,
        "Content-Type": "application/json"
    }
    json_payload = json.dumps({
        "q": query,
        "gl": 'cn' if target_lang == 'zh' else 'us',
        'hl': 'zh-cn' if target_lang == 'zh' else 'en',
    })

    async with httpx.AsyncClient(proxy=G.A.PROXY_ADDR) as client:
        response = await client.post(url, headers=headers, data=json_payload, timeout=30)
        response_json = response.json()

    results_formatted = [{"title": it['title'], "text": it['snippet']} for it in response_json['organic'] if 'snippet' in it]
    return results_formatted

if __name__ == "__main__":
    async def main():
        from maica import init
        init()
        print(await asearch("今天的新闻"))

    asyncio.run(main())
    


import json
from dataclasses import dataclass

from maica.maica_utils import *

@dataclass
class TpAPIKeys():...

def pkg_init_api_keys():
    api_keys: dict = json.loads(G.A.TP_APIS)
    for k, v in api_keys.items():
        setattr(TpAPIKeys, k, v)

if __name__ == "__main__":
    pkg_init_api_keys()


"""Import layer 2.5"""
from typing import *
from dataclasses import dataclass
from websockets import ServerConnection
from Crypto.Random import random as crandom
from .setting_utils import MaicaSettings

class TracerayId():
    """String-like."""
    var: str=''
    def __init__(self):
        self.var = str(crandom.randint(0,9999999999)).zfill(10)
    def rotate(self):
        self.__init__()
    def __str__(self):
        return self.var

@dataclass
class RealtimeSocketsContainer():
    """For no-setting usage."""
    websocket: Optional[ServerConnection]=None
    traceray_id: TracerayId=None
    maica_settings: MaicaSettings=None

    def __post_init__(self):
        if not self.traceray_id:
            self.traceray_id = TracerayId()
        if not self.maica_settings:
            self.maica_settings = MaicaSettings()

    def rotate_tid(self):
        """Rotate traceray_id."""
        self.traceray_id = TracerayId()


"""
Some annoying migration stuff here.
"""
import asyncio
import importlib
from typing import *
from packaging.version import parse
from collections.abc import Callable
from maica.maica_utils import *
from maica.initializer import create_marking

available_list: list[tuple[str, Callable]]

def pkg_init_migrations():
    global available_list
    available_list = []

    prio = 0
    while True:
        prio += 1
        modname = f".migration_{prio}"
        try:
            mig_module = importlib.import_module(modname, "maica.initializer.migrations")
            mig_v = parse(mig_module.upper_version); migrate = mig_module.migrate
            available_list.append((mig_v, migrate))
        except Exception: break

    sync_messenger(info=f'[maica-mig] {len(available_list)} migrations found', type=MsgType.DEBUG)

def migrate(version):
    migrated = False

    last_version_parsed = parse(version)
    curr_version = load_env('MAICA_CURR_VERSION')
    curr_version_parsed = parse(curr_version)

    sync_messenger(info=f'[maica-mig] Last initialized version {version}, current version {curr_version}', type=MsgType.DEBUG)

    for mig_item in available_list:
        if curr_version_parsed >= mig_item[0] > last_version_parsed:
            try:
                sync_messenger(info=f'Running migration upper-version {str(mig_item[0])}...', type=MsgType.PRIM_LOG)
                asyncio.run(mig_item[1]())
                sync_messenger(info=f'Finished migration upper-version {str(mig_item[0])}', type=MsgType.PRIM_LOG)
                migrated = True
            except CommonMaicaException as ce:
                if ce.is_critical:
                    raise ce
                else:
                    sync_messenger(error=ce, no_raise=True)
                    migrated = True
    if migrated:
        sync_messenger(info=f'[maica-mig] Migration finished, continuing launch procedure...', type=MsgType.LOG)
    else:
        sync_messenger(info=f'[maica-mig] No migration applied, continuing launch procedure...', type=MsgType.DEBUG)

    return migrated



import asyncio
import datetime
import os

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.interval import IntervalTrigger

from typing import *
from maica.mtools import ProcessingImg
from maica.maica_utils import *

_CONNS_LIST = ['maica_pool']

class CommonScheduler():
    """Keeps a schedule running."""

    root_csc: ConnSocketsContainer = None
    """Don't forget to implement at first!"""

    def __init__(self, involve_chat=True, involve_tts=True):
        rsc = RealtimeSocketsContainer()
        csc = self.__class__.root_csc.spawn_sub(rsc)
        self.fsc = FullSocketsContainer(rsc, csc)
        self.maica_pool = self.fsc.maica_pool

        self.schedule = AsyncIOScheduler()
        if involve_chat:
            self.schedule.add_job(
                self.rotate_ms_cache,
                trigger=IntervalTrigger(hours=1),
                id='rotate_ms_cache',
            )
            self.schedule.add_job(
                self.rotate_mv_imgs,
                trigger=IntervalTrigger(hours=1),
                id='rotate_mv_imgs'
            )
        if involve_tts:
            self.schedule.add_job(
                self.rotate_mtts_cache,
                trigger=IntervalTrigger(hours=1),
                id='rotate_mtts_cache'
            )
        self.schedule.start()

    @Decos.log_task
    async def rotate_ms_cache(self):
        """Deletes outdated mscache. We usually want to keep them for trainings."""
        keep_time = int(G.A.ROTATE_MSCACHE)
        if keep_time:
            timestamp = datetime.datetime.now()
            earliest_timestamp = timestamp - datetime.timedelta(hours=keep_time)
            sql_expression_1 = "DELETE FROM ms_cache WHERE timestamp < %s"
            rows = (await self.maica_pool.query_modify(expression=sql_expression_1, values=(earliest_timestamp, )))[0]
            sync_messenger(info=f'Removed {rows} rows of MSpire cache', type=MsgType.LOG)

    @Decos.log_task
    async def rotate_mv_imgs(self):
        """Deletes outdated mv_img."""
        keep_time = int(G.A.ROTATE_MVISTA)
        if keep_time:
            timestamp = datetime.datetime.now()
            earliest_timestamp = timestamp - datetime.timedelta(hours=keep_time)

            sql_expression_1 = "SELECT uuid FROM mv_meta WHERE timestamp < %s"
            result = await self.maica_pool.query_get(expression=sql_expression_1, values=(earliest_timestamp, ), fetchall=True)
            uuids = [i[0] for i in result]
            for uuid in uuids:

                processing_img = ProcessingImg()
                processing_img.det_path(uuid)
                processing_img.delete()

                sql_expression_2 = "DELETE FROM mv_meta WHERE uuid = %s"
                await self.maica_pool.query_modify(expression=sql_expression_2, values=(uuid, ))

            sync_messenger(info=f'Removed {len(uuids)} MVista images', type=MsgType.LOG)

    @Decos.log_task
    async def rotate_mtts_cache(self):
        """Deletes long-unused mtts cache."""
        def sync_rotation(earliest_timestamp):
            base_path = get_inner_path('fs_storage/mtts')
            mtts_cache_entries = os.scandir(base_path)
            delete_list = []; all_count = 0
            for entry in mtts_cache_entries:
                if entry.is_file() and not entry.name.startswith('.'):
                    all_count += 1
                    path = entry.path
                    last_atime = os.path.getatime(path)
                    if last_atime < earliest_timestamp:
                        delete_list.append(path)
            
            for path in delete_list:
                try:
                    os.remove(path)
                except Exception:...

            return len(delete_list), all_count

        keep_time = int(G.T.ROTATE_TTSCACHE)
        if keep_time:
            timestamp = datetime.datetime.now()
            earliest_timestamp_float = (timestamp - datetime.timedelta(hours=keep_time)).timestamp()
            deleted_count, all_count = await wrap_run_in_exc(None, sync_rotation, earliest_timestamp_float)

            sync_messenger(info=f'Removed {deleted_count} MTTS caches, {all_count} remaining', type=MsgType.LOG)

    async def run_schedule(self):
        """The schedule starter."""
        await sleep_forever()

    async def close(self):
        self.schedule.shutdown()

async def prepare_thread(**kwargs):

    # Construct csc first
    root_csc_kwargs = {k: kwargs.get(k) for k in _CONNS_LIST}
    root_csc = ConnSocketsContainer(**root_csc_kwargs)
    CommonScheduler.root_csc = root_csc

    await messenger(info='MAICA scheduler started!', type=MsgType.PRIM_SYS)
    
    try:
        common_scheduler = CommonScheduler(kwargs.get('involve_chat', True), kwargs.get('involve_tts', True))
        await common_scheduler.run_schedule()
        common_scheduler.close()

    except BaseException as be:
        if isinstance(be, Exception):
            error = CommonMaicaError(str(be), '504')
            await messenger(error=error, no_raise=True)
    finally:
        await messenger(info='MAICA scheduler stopped!', type=MsgType.PRIM_SYS)

async def _run_shd():
    """
    Notice: these only happen running individually!
    Use prepare_thread() for lower level control.
    """
    from maica import init
    init()
    _root_csc_items = [getattr(ConnUtils, k)() for k in _CONNS_LIST]
    root_csc_items = await asyncio.gather(*_root_csc_items)
    root_csc_kwargs = dict(zip(_CONNS_LIST, root_csc_items))

    task = asyncio.create_task(prepare_thread(**root_csc_kwargs))
    await task

    close_list = []
    for conn in root_csc_items:
        close_list.append(conn.close())
    await asyncio.gather(*close_list)
    await messenger(info='Individual MAICA scheduler cleaning done', type=MsgType.DEBUG)

def run_shd():
    asyncio.run(_run_shd())

if __name__ == '__main__':
    run_shd()



"""
SERP is really complex if you don't use google's super expensive api, so I made this.

This searches submodules in its folder at a priority sequence from 0 to 99, and adopts 
the first module that implements a valid 'asearch' function with optional 'requires'.

'asearch' must accept (query, target_lang), and returns a list[dict[#type, #content]]. This 
implementation will only take effect if all items in 'requires' exist in TpAPIKeys.

The local solution is not stable and efficient enough, so I might be using some third-
party apis. They're not ads.

You can always implement your own SERP function by adding a submodule with higher priority 
(smaller number).
"""
import importlib
import random
from typing import *
from collections.abc import Callable
from maica.maica_utils import *
from maica.mtools.api_keys import TpAPIKeys

@overload
async def asearch(query: str, target_lang: Literal['zh', 'en']) -> list[dict[Literal['title', 'text'], Annotated[str, Desc('Content')]]]:...

available_list: list[tuple[int, Callable]]
last_used = -1

def pkg_init_serp_provider():
    global available_list
    serp_provider = None; asearch = None; requires = None
    available_list = []

    if G.A.NO_SERP != '1':
        for prio in range(0, 100):
            modname = f".serp_provider_{prio}"
            try:
                serp_provider = importlib.import_module(modname, "maica.mtools.providers")
                asearch = serp_provider.asearch; requires = serp_provider.requires
                for r in requires:
                    assert getattr(TpAPIKeys, r)
                available_list.append((prio, asearch))
            except Exception:...
    sync_messenger(info=f"[maica-serp] Available SERP providers: {', '.join([str(i[0]) for i in available_list]) or None}", type=MsgType.DEBUG)

def get_asearch(avoid: Union[Literal['last'], int]=None, rand: bool=False) -> Callable:
    global last_used

    if avoid == 'last':
        avoid = last_used

    avail_temp = [e for e in available_list if e[0] != avoid]
    if not avail_temp:
        avail_temp = available_list
    if rand:
        selected = random.choice(avail_temp) if avail_temp else (-1, None)
    else:
        selected = avail_temp[0] if avail_temp else (-1, None)

    last_used = selected[0]; asearch = selected[1]

    return asearch

# Normally you should not do 'from provider import ...' to keep it dynamic
__all__ = ['get_asearch']


"""Import layer 3"""
import aiomysql
import aiosqlite
import asyncio
import traceback
import json

from typing import *
from typing_extensions import deprecated
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion
from .gvars import *
from .maica_utils import *
from .setting_utils import *
from .fsc_early import *
from .locater import *

def pkg_init_connection_utils():
    global ConnUtils

    if G.A.DB_ADDR != "sqlite":
        """We suppose we're using MySQL."""
        async def auth_pool(ro=True):
            return await DbPoolManager.async_create(
                host=G.A.DB_ADDR,
                db=G.A.AUTH_DB,
                user=G.A.DB_USER,
                password=G.A.DB_PASSWORD,
                ro=ro,
            )

        async def maica_pool(ro=False):
            return await DbPoolManager.async_create(
                host=G.A.DB_ADDR,
                db=G.A.DATA_DB,
                user=G.A.DB_USER,
                password=G.A.DB_PASSWORD,
                ro=ro,
            )
        
        async def basic_pool(ro=False):
            return await DbPoolManager.async_create(
                host=G.A.DB_ADDR,
                db=None,
                user=G.A.DB_USER,
                password=G.A.DB_PASSWORD,
                ro=ro,
            )
    else:
        """We suppose we're using SQLite."""
        async def auth_pool(ro=True):
            return await SqliteDbPoolManager.async_create(
                db=get_inner_path(G.A.AUTH_DB)
            )
        
        async def maica_pool(ro=False):
            return await SqliteDbPoolManager.async_create(
                db=get_inner_path(G.A.DATA_DB)
            )
        
        async def basic_pool(ro=False):
            """
            There's no host concept in SQLite, so no basic pool.
            """
            return None
    setattr(ConnUtils, 'auth_pool', auth_pool)
    setattr(ConnUtils, 'maica_pool', maica_pool)
    setattr(ConnUtils, 'basic_pool', basic_pool)

def test_logger(func):
    async def wrapper(self, expression, values, *args, **kwargs):
        print(f'Query: {expression}\nValues: {values}')
        result = await func(self, expression, values, *args, **kwargs)
        print(f'Result: {result}')
        return result
    return wrapper

def apply_postfix(messages, thinking: Literal[True, False, None]=None):
    last_msg: dict = messages[-1]
    if last_msg.get('role') == 'user':
        if isinstance(last_msg.get('content'), str):
            match thinking:
                case True:
                    last_msg['content'] += G.A.MFOCUS_THINK
                case False:
                    last_msg['content'] += G.A.MFOCUS_NOTHINK
        elif isinstance(last_msg.get('content'), list):
            d: dict
            for d in last_msg['content']:
                if d.get('type') == 'text' and isinstance(d.get('text'), str):
                    match thinking:
                        case True:
                            d['text'] += G.A.MFOCUS_THINK
                        case False:
                            d['text'] += G.A.MFOCUS_NOTHINK
                    break
        else:
            raise MaicaInputError('Context schedule not recognizable')
    return messages

class DbPoolManager(AsyncCreator):
    """Maintain a database connection pool so you don't have to."""

    db_type: Literal['mysql', 'sqlite'] = 'mysql'

    def __init__(self, db, host, user, password, ro=False):
        self.db, self.host, self.user, self.password, self.ro = db, host, user, password, ro
        self.name = self.db
        self.pool: aiomysql.Pool = None
        self.pool_container: list[aiomysql.Pool] = []
        """We use this thing to sync sub-instances' pools with mother instance."""
        self.lock = asyncio.Lock()
        """Pool affecting actions are performed with lock acquired."""

    @Decos.catch_exceptions
    async def _ainit(self):
        """Initialize MySQL connection pool."""
        if not self.lock.locked():
            async with self.lock:
                await self.close()
                self.pool = await aiomysql.create_pool(host=self.host, user=self.user, password=self.password, db=self.db)
                self.pool_container.append(self.pool)
        else:
            async with self.lock:
                return

    @Decos.catch_exceptions
    async def keep_alive(self):
        """Check and maintain MySQL connection."""
        try:
            async with self.pool.acquire():
                pass
            return
        except Exception:
            await messenger(info=f"Recreating {self.name} pool since cannot acquire", type=MsgType.WARN)
            await self._ainit()

    @overload
    async def query_get(self, expression: str, values: Optional[tuple]=None, fetchall: bool=False, inherit_conn: Optional[aiomysql.Connection]=None) -> tuple:
        """Execute SELECT query on MySQL database."""

    # @test_logger
    @Decos.catch_exceptions
    @Decos.ro_expression
    @Decos.conn_retryer_factory()
    async def query_get(self, expression, values=None, fetchall=False, inherit_conn: Optional[aiomysql.Connection]=None) -> tuple:
        async def _query_get(cur, expression, values, fetchall) -> tuple:
            if not values:
                await cur.execute(expression)
            else:
                await cur.execute(expression, values)
            results = await cur.fetchone() if not fetchall else await cur.fetchall()
            return results

        results = None
        await self.keep_alive()

        if not inherit_conn:
            async with self.pool.acquire() as conn:
                async with conn.cursor() as cur:
                    results = await _query_get(cur, expression, values, fetchall)
                    
                    # aiomysql caches results in some cases if you don't commit
                    # Really weird to me
                    await conn.commit()
        else:
            async with inherit_conn.cursor() as cur:
                results = await _query_get(cur, expression, values, fetchall)
                # We leave the transactional manager handling commitments

        return results

    # @test_logger
    @Decos.catch_exceptions
    @Decos.wo_expression
    @Decos.conn_retryer_factory()
    async def query_modify(self, expression: str, values: Optional[tuple]=None, fetchall=False, inherit_conn: Optional[aiomysql.Connection]=None) -> tuple[Annotated[int, Desc('rows')], Annotated[int, Desc('lrid')]]:
        """Execute INSERT/UPDATE/DELETE query on MySQL database."""
        if self.ro:
            raise MaicaDbError(f'DB marked as ro, no modification permitted', '511', 'db_modification_denied')
        
        async def _query_modify(cur, expression, values, fetchall) -> tuple[int, int]:
            if not values:
                rows = await cur.execute(expression)
            else:
                rows = await cur.execute(expression, values)
            lrid = cur.lastrowid
            return rows, lrid

        await self.keep_alive()

        if not inherit_conn:
            async with self.pool.acquire() as conn:
                async with conn.cursor() as cur:
                    res = await _query_modify(cur, expression, values, fetchall)
                    await conn.commit()
        else:
            async with inherit_conn.cursor() as cur:
                res = await _query_modify(cur, expression, values, fetchall)
                # We leave the transactional manager handling commitments

        return res
    
    async def close(self):
        """Close before closing coroutine to avoid errors."""
        try:
            self.pool.close()
            await self.pool.wait_closed()
        except Exception:...
        finally:
            self.pool_container.clear()

    def summon_sub(self, rsc: Optional[RealtimeSocketsContainer]=None):
        """Summons a per-user instance."""
        return SubDbPoolManager(self, rsc)

class SubDbPoolManager(DbPoolManager):
    """Per-user DbPoolManager."""
    def __init__(self, parent: DbPoolManager, rsc: Optional[RealtimeSocketsContainer]=None):
        """Must summon from a parent object."""
        for k, v in vars(parent).items():
            setattr(self, k, v)

        self.parent = parent; self.rsc = rsc

    async def _ainit(self):
        """Do not use async_create."""
        raise NotImplementedError
    
    async def keep_alive(self):
        await self.parent.keep_alive()
        self.pool = self.pool_container[0]

    async def close(self):...

class SqliteDbPoolManager(DbPoolManager):
    """
    SQLite-specific database pool coroutine.
    Notice: since SQLite only allow rollback for TRANSACTION which differs
    from MySQL, we're not implementing transaction management for SQLite for
    now. 
    """

    db_type: Literal['mysql', 'sqlite'] = 'sqlite'

    def __init__(self, db, host=None, user=None, password=None, ro=False):
        self.db_path = db
        self.name = self.db_path
        self.ro = ro
        self.pool: aiosqlite.Connection = None
        self.pool_container: list[aiosqlite.Connection] = []
        self.lock = asyncio.Lock()

    @Decos.catch_exceptions
    async def _ainit(self):
        """
        Initialize SQLite connection pool.
        Notice: there is actually no 'pool' concept in SQLite, but
        we're considering connections as pools to keep symmetry with
        the corresponding MySQL functions.
        """
        if not self.lock.locked():
            async with self.lock:
                await self.close()
                self.pool = aiosqlite.connect(self.db_path)
                await self.pool.__aenter__()
                if self.ro:
                    await self.pool.execute("PRAGMA query_only = ON")
                self.pool_container.append(self.pool)
        else:
            async with self.lock:
                return

    @Decos.catch_exceptions
    async def keep_alive(self):
        """Check and maintain SQLite connection."""
        try:
            await self.pool.execute("SELECT 1")
        except Exception:
            await messenger(info=f"Recreating {self.db} pool since cannot acquire", type=MsgType.WARN)
            await self._ainit()

    @overload
    async def query_get(self, expression: str, values: Optional[tuple]=None, fetchall: bool=False) -> tuple:
        """Execute SELECT query on SQLite database."""

    # @test_logger
    @Decos.catch_exceptions
    @Decos.ro_expression
    @Decos.escape_sqlite_expression
    @Decos.conn_retryer_factory()
    async def query_get(self, expression, values=None, fetchall=False) -> tuple:
        results = None
        await self.keep_alive()

        if not values:
            cursor = await self.pool.execute(expression)
        else:
            cursor = await self.pool.execute(expression, values)
        results = await cursor.fetchone() if not fetchall else await cursor.fetchall()
        await self.pool.commit()

        return results

    # @test_logger
    @Decos.catch_exceptions
    @Decos.wo_expression
    @Decos.escape_sqlite_expression
    @Decos.conn_retryer_factory()
    async def query_modify(self, expression: str, values: Optional[tuple]=None, fetchall=False) -> tuple[int, int]:
        """Execute INSERT/UPDATE/DELETE query on SQLite database."""
        if self.ro:
            raise MaicaDbError(f'DB marked as ro, no modification permitted', '511', 'sqlite_modification_denied')
        lrid = None

        await self.keep_alive()
        if not values:
            cursor = await self.pool.execute(expression)
        else:
            cursor = await self.pool.execute(expression, values)
        await self.pool.commit()
        rows = cursor.rowcount
        lrid = cursor.lastrowid
        await cursor.close()

        return rows, lrid

    async def close(self):
        """Close SQLite connection."""
        try:
            await self.pool.close()
        except Exception:...
        finally:
            self.pool_container.clear()

    def summon_sub(self, rsc: Optional[RealtimeSocketsContainer]=None):
        """Summons a per-user instance."""
        return SubSqliteDbPoolManager(self, rsc)

class SubSqliteDbPoolManager(SqliteDbPoolManager):
    """Per-user SqliteDbPoolManager."""
    def __init__(self, parent: SqliteDbPoolManager, rsc: Optional[RealtimeSocketsContainer]=None):
        """Must summon from a parent object."""
        for k, v in vars(parent).items():
            setattr(self, k, v)

        self.parent = parent; self.rsc = rsc

    async def _ainit(self):
        """Do not use async_create."""
        raise NotImplementedError
    
    async def keep_alive(self):
        await self.parent.keep_alive()
        self.pool = self.pool_container[0]

    async def close(self):...

class AiConnectionManager(AsyncCreator):
    """Maintain an AI connection so you don't have to."""
    def __init__(self, api_key, base_url, name='ai_conn', model: Union[int, str]=0):
        self.test = False
        self.api_key, self.base_url, self.name, self.model = api_key, base_url, name, model
        self.gen_kwargs = {}
        self.sock_container: list[AsyncOpenAI, str] = []
        """We use this thing to sync sub-instances' socks with mother instance."""
        self.lock = asyncio.Lock()
        """Socket affecting actions are performed with lock acquired."""

    @Decos.catch_exceptions
    async def _ainit(self):
        if not self.base_url:
            self.test = True
            return
        else:
            if not self.lock.locked():
                async with self.lock:
                    await self.close()
                    self._open_socket()
                    await self._select_model()
            else:
                async with self.lock:
                    return

    def _open_socket(self):
        self.socket = AsyncOpenAI(api_key=self.api_key, base_url=self.base_url)
        self.sock_container.append(self.socket)

    async def _select_model(self):
        model_list = await self.socket.models.list()
        models = model_list.data
        if isinstance(self.model, int):
            self.model_actual = models[0].id
        else:
            self.model_actual = self.model
        self.sock_container.append(self.model_actual)

    def default_params(self, **kwargs):
        """These params will always be applied to generations. Overwritten."""
        self.gen_kwargs = kwargs

    @Decos.catch_exceptions
    async def keep_alive(self):
        """Check and maintain OpenAI connection."""
        if self.socket.is_closed():
            await messenger(info=f"Recreating {self.name} sock since is closed", type=MsgType.WARN)
            await self._ainit()

    @Decos.catch_exceptions
    @Decos.conn_retryer_factory()
    async def make_completion(self, **kwargs) -> ChatCompletion:
        """Makes completion with arguments."""
        kwargs.update(
            {
                "model": self.model_actual
            }
        )
        mixed_exbody = {**self.gen_kwargs.get('extra_body', {}), **kwargs.get('extra_body', {})}
        mixed_kwargs = {**self.gen_kwargs, **kwargs}
        mixed_kwargs['extra_body'] = mixed_exbody

        await self.keep_alive()
        task_stream_resp = asyncio.create_task(self.socket.chat.completions.create(**mixed_kwargs))
        await asyncio.wait_for(task_stream_resp, timeout=int(G.A.OPENAI_TIMEOUT) if G.A.OPENAI_TIMEOUT != '0' else None)
        return task_stream_resp.result()
    
    async def close(self):
        try:
            await self.socket.close()
        except Exception:...
        finally:
            self.sock_container.clear()

    def summon_sub(self, rsc: Optional[RealtimeSocketsContainer]=None):
        """Summons a per-user instance."""
        return SubAiConnectionManager(self, rsc)
    
class SubAiConnectionManager(AiConnectionManager):
    """Per-user AiConnectionManager."""
    def __init__(self, parent: AiConnectionManager, rsc: Optional[RealtimeSocketsContainer]=None):
        """Must summon from a parent object."""
        for k, v in vars(parent).items():
            setattr(self, k, v)

        self.parent = parent; self.rsc = rsc

    async def _ainit(self):
        """Do not use async_create."""
        raise NotImplementedError
    
    async def keep_alive(self):
        await self.parent.keep_alive()
        self.socket, self.model_actual = self.parent.sock_container

    async def close(self):...

class ConnUtils():
    """Just a wrapping for functions."""
    async def auth_pool(ro=True) -> DbPoolManager:
        """Dummy."""
    async def maica_pool(ro=False) -> DbPoolManager:
        """Dummy."""
    async def basic_pool(ro=False) -> DbPoolManager:
        """Dummy."""

    async def mcore_conn():
        conn = await AiConnectionManager.async_create(
            api_key=G.A.MCORE_KEY,
            base_url=G.A.MCORE_ADDR,
            name='mcore_conn',
            model=G.A.MCORE_CHOICE or 0,
        )
        conn.default_params(**json.loads(G.A.MCORE_EXTRA))
        return conn

    async def mfocus_conn():
        conn = await AiConnectionManager.async_create(
            api_key=G.A.MFOCUS_KEY,
            base_url=G.A.MFOCUS_ADDR,
            name='mfocus_conn',
            model=G.A.MFOCUS_CHOICE or 0,
        )
        conn.default_params(**json.loads(G.A.MFOCUS_EXTRA))
        return conn
    
    async def mvista_conn():
        """Disable if no addr provided."""
        if G.A.MVISTA_ADDR:
            conn = await AiConnectionManager.async_create(
                api_key=G.A.MVISTA_KEY,
                base_url=G.A.MVISTA_ADDR,
                name='mvista_conn',
                model=G.A.MVISTA_CHOICE or 0,
            )
            conn.default_params(**json.loads(G.A.MVISTA_EXTRA))
            return conn
        else:
            return None

    async def mnerve_conn():
        """Disable if no addr provided."""
        if G.A.MNERVE_ADDR:
            conn = await AiConnectionManager.async_create(
                api_key=G.A.MNERVE_KEY,
                base_url=G.A.MNERVE_ADDR,
                name='mnerve_conn',
                model=G.A.MNERVE_CHOICE or 0,
            )
            conn.default_params(**json.loads(G.A.MNERVE_EXTRA))
            return conn
        else:
            return None

async def validate_input(input: Union[str, dict, list], limit: int=0, rsc: Optional[RealtimeSocketsContainer]=None, must: Optional[list]=None, warn: Optional[list]=None) -> Union[dict, list]:
    """
    Mostly for ws.
    """
    must = must or []
    warn = warn or []
    if not input:
        raise MaicaInputWarning('Input is empty', '410', 'maica_input_empty')
    
    if isinstance(input, str):
        if limit and len(input) > limit:
            raise MaicaInputWarning('Input length exceeded', '413', 'maica_input_length_exceeded')
        try:
            input_json = json.loads(input)
        except Exception as e:
            raise MaicaInputWarning('Request body not JSON', '400', 'maica_input_not_json') from e
    elif isinstance(input, dict | list):
        if limit and len(str(input)) > limit:
            raise MaicaInputWarning('Input length exceeded', '413', 'maica_input_length_exceeded')
        input_json = input
    else:
        raise MaicaInputError('Input must be string or JSON-like', '400', 'maica_input_validation_denied')

    if must:
        for mustkey in must:
            if input_json.get(mustkey) is None:
                raise MaicaInputWarning(f'Request contains no necessary {mustkey}', '405', 'maica_input_necessity_missing')
    if warn:
        for warnkey in warn:
            if input_json.get(warnkey) is None:
                if rsc:
                    await messenger(rsc.websocket, 'maica_future_warning', f'Requests containing no {warnkey} will likely be deprecated in the future', '302', type=MsgType.WARN)
    
    return input_json



import os
from PIL import Image
from io import BytesIO
import magic

import uuid
from typing import *
from maica.maica_utils import *

_ALLOWED_MIMES = ['image/jpeg', 'image/png', 'image/bmp', 'image/webp']

class ProcessingImg():
    """A wrapped picture to process."""
    _bio: Optional[BytesIO] = None
    _real_path: Optional[str] = None
    _base_path: str = get_inner_path('fs_storage')

    format: str = ''
    is_compressed: bool = False

    @property
    def file_name(self):
        return os.path.basename(self._real_path)

    def _rfb(self, *args, **kwargs):
        self._bio.seek(0)
        b = self._bio.read(*args, **kwargs)
        self._bio.seek(0)
        return b
    
    def _purge(self):
        self._bio.seek(0)
        self._bio.truncate(0)

    def _save(self, path):
        self._bio.seek(0)
        with open(path, 'wb') as f:
            f.write(self._bio.getvalue())

    def _read(self, path):
        self._purge()
        with open(path, 'rb') as f:
            self._bio.write(f.read())

    @overload
    def __init__(self, input: bytes) -> None:
        """Create from binary."""

    @overload
    def __init__(self, input: str) -> None:
        """Read from fs."""

    @overload
    def __init__(self) -> None:
        """Do whatever later manually."""

    def __init__(self, input=None):
        """Automatically select initialization method."""
        self._bio = BytesIO()
        if isinstance(input, str):
            self.extract(input)
        elif input:
            self.perfuse(input)

    def perfuse(self, binary: bytes):
        """Create from binary."""
        self._bio.write(binary)
        self.format = magic.from_buffer(self._rfb(2048), mime=True)
        if not self.format in _ALLOWED_MIMES:
            raise MaicaInputWarning(f'Input file {self.format} is not image')
        self.compress()

    def extract(self, uuid: str):
        """Read from fs."""
        self.det_path(uuid)
        self.read()
        self.format = magic.from_buffer(self._rfb(2048), mime=True)
        if not self.format in _ALLOWED_MIMES:
            raise MaicaInputError(f'Extracted file {self.format} is not image')
        
    def compress(self, mw=1920, mh=1080, q=85):
        """Compresses the wrapped picture."""
        img = Image.open(self._bio)
        img.load()

        if img.mode in ('RGBA', 'LA', 'P'):
            background = Image.new('RGB', img.size, (255, 255, 255))
            if img.mode == 'P':
                img = img.convert('RGBA')
            background.paste(img, mask=img.split()[-1] if img.mode == 'RGBA' else None)
            img = background
        elif img.mode != 'RGB':
            img = img.convert('RGB')

        width, height = img.size
        scale = min(mw / width, mh / height, 1.0)

        if scale < 1.0:
            new_size = (int(width * scale), int(height * scale))
            img = img.resize(new_size, Image.Resampling.LANCZOS)

        self._purge()
        img.save(self._bio, format='JPEG', quality=q, optimize=True)
        self.is_compressed = True

    def det_path(self, fname: Optional[str]=None):
        """Just fill the _real_path."""
        def abspath():
            nonlocal fname
            try:
                fname = str(uuid.UUID(fname))
            except Exception as e:
                raise MaicaInputWarning(f'fname {fname} is not uuid')
            return os.path.join(self._base_path, f"{fname}.jpg")
        
        if not fname:
            fname = str(uuid.uuid4())
            while os.path.isfile(abspath()):
                fname = str(uuid.uuid4())
        self._real_path = abspath()
        return fname

    def read(self):
        """Read bytes from desired path."""
        assert os.path.isfile(self._real_path), f"File {self.file_name} not exist"
        self._read(self._real_path)

    def save(self):
        """Save wrapped picture to desired path."""
        assert self._real_path, "Run det_path before save"
        self._save(self._real_path)

    def delete(self):
        """Purge bio and delete desired path."""
        self._purge()
        assert os.path.isfile(self._real_path), f"File {self.file_name} not exist"
        os.remove(self._real_path)

    def to_bio(self):
        """Get stored bio."""
        return self._bio
